import argparse
import asyncio
import base64
import datetime
import hashlib
import html
import logging
import re
import ssl
import tempfile
from argparse import Namespace
from base64 import b32encode
from collections import defaultdict
from time import time
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import irc.client
import irc.client_aio
import irc.connection
from jaraco.stream import buffer
from mautrix.util.bridge_state import BridgeStateEvent
from python_socks.async_.asyncio import Proxy

from heisenbridge import __version__
from heisenbridge.channel_room import ChannelRoom
from heisenbridge.command_parse import CommandManager
from heisenbridge.command_parse import CommandParser
from heisenbridge.command_parse import CommandParserError
from heisenbridge.irc import HeisenReactor
from heisenbridge.plumbed_room import PlumbedRoom
from heisenbridge.private_room import parse_irc_formatting
from heisenbridge.private_room import PrivateRoom
from heisenbridge.private_room import unix_to_local
from heisenbridge.room import Room
from heisenbridge.space_room import SpaceRoom


def connected(f):
    def wrapper(*args, **kwargs):
        self = args[0]

        if not self.conn or not self.conn.connected:
            self.send_notice("Need to be connected to use this command.")
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


# forwards events to private and channel rooms
def ircroom_event(target_arg=None):
    def outer(f):
        def wrapper(self, conn, event):
            if target_arg is not None:
                # if we have target arg use that
                target = event.arguments[target_arg].lower()
            else:
                # switch target around if it's targeted towards us directly
                target = event.target.lower() if event.target != conn.real_nickname else event.source.nick.lower()

            if target in self.rooms:
                room = self.rooms[target]
                try:
                    room_f = getattr(room, "on_" + event.type)
                    try:
                        return room_f(conn, event)
                    except Exception:
                        logging.exception(f"Calling on_{event.type} failed for {target}")
                except AttributeError:
                    logging.warning(f"Expected {room} to have on_{event.type} but didn't")

            return f(self, conn, event)

        return wrapper

    return outer


class NetworkRoom(Room):
    # configuration stuff
    name: str
    connected: bool
    nick: str
    username: str
    ircname: str
    password: str
    sasl_mechanism: str
    sasl_username: str
    sasl_password: str
    autocmd: str
    pills_length: int
    pills_ignore: list
    autoquery: bool
    color: bool
    tls_cert: str
    rejoin_invite: bool
    rejoin_kick: bool
    caps: list
    forward: bool

    # state
    commands: CommandManager
    conn: Any
    rooms: Dict[str, Room]
    connecting: bool
    real_host: str
    real_user: str
    pending_kickbans: Dict[str, List[Tuple[str, str]]]
    backoff: int
    backoff_task: Any
    next_server: int
    connected_at: int
    space: SpaceRoom
    post_init_done: bool
    caps_supported: list
    caps_enabled: list
    caps_task: Any

    def init(self):
        self.name = None
        self.connected = False
        self.nick = None
        self.username = None
        self.ircname = None
        self.password = None
        self.sasl_mechanism = None
        self.sasl_username = None
        self.sasl_password = None
        self.autocmd = None
        self.pills_length = 2
        self.pills_ignore = []
        self.autoquery = True
        self.color = True
        self.allow_ctcp = False
        self.tls_cert = None
        self.rejoin_invite = True
        self.rejoin_kick = False
        self.caps = ["message-tags", "chghost", "znc.in/self-message"]
        self.forward = False
        self.backoff = 0
        self.backoff_task = None
        self.next_server = 0
        self.connected_at = 0

        self.commands = CommandManager()
        self.conn = None
        self.rooms = {}
        self.connlock = asyncio.Lock()
        self.disconnect = True
        self.real_host = "?" * 63  # worst case default
        self.real_user = "?" * 8  # worst case default
        self.keys = {}  # temp dict of join channel keys
        self.keepnick_task = None  # async task
        self.whois_data = defaultdict(dict)  # buffer for keeping partial whois replies
        self.pending_kickbans = defaultdict(list)
        self.space = None
        self.post_init_done = False
        self.caps_supported = []
        self.caps_enabled = []
        self.caps_task = None

        cmd = CommandParser(
            prog="NICK",
            description="set/change nickname",
            epilog=(
                "You can always see your current nickname on the network without arguments.\n"
                "If connected new nickname will be sent to the server immediately. It may be rejected and an underscore appended"
                " to it automatically.\n"
            ),
        )
        cmd.add_argument("nick", nargs="?", help="new nickname")
        self.commands.register(cmd, self.cmd_nick)

        cmd = CommandParser(
            prog="USERNAME",
            description="set username",
            epilog=(
                "Setting a new username requires reconnecting to the network.\n"
                "\n"
                "Note: If identd is enabled this will be ignored and Matrix ID hash or admin set custom ident is used."
            ),
        )
        cmd.add_argument("username", nargs="?", help="new username")
        cmd.add_argument("--remove", action="store_true", help="remove stored username")
        self.commands.register(cmd, self.cmd_username)

        cmd = CommandParser(
            prog="REALNAME",
            description="set realname",
            epilog=("Setting a new realname requires reconnecting to the network.\n"),
        )
        cmd.add_argument("name", nargs="?", help="new realname")
        cmd.add_argument("--remove", action="store_true", help="remove stored name")
        self.commands.register(cmd, self.cmd_realname)

        cmd = CommandParser(
            prog="PASSWORD",
            description="set server password",
            epilog=(
                "You can store your network password using this command and it will be automatically offered on connect.\n"
                "Some networks allow using this to identify with NickServ on connect without sending a separate message.\n"
                "\n"
                "Note: Bridge administrators can trivially see the stored password if they want to.\n"
            ),
        )
        cmd.add_argument("password", nargs="?", help="new password")
        cmd.add_argument("--remove", action="store_true", help="remove stored password")
        self.commands.register(cmd, self.cmd_password)

        cmd = CommandParser(
            prog="SASL",
            description="set SASL PLAIN credentials",
            epilog=(
                "If the network supports SASL authentication you can configure them with this command.\n"
                "If your password contains special characters, use shell escaping.\n"
                'Example: SASL "pass;word"\n'
                "\n"
                "Note: Bridge administrators can trivially see the stored password if they want to.\n"
            ),
        )
        cmd.add_argument(
            "--mechanism", choices=["auto", "none", "plain", "external"], help="SASL authentication mechanism"
        )
        cmd.add_argument("--username", help="SASL username")
        cmd.add_argument("--password", help="SASL password")
        cmd.add_argument("--remove", action="store_true", help="remove stored credentials")
        self.commands.register(cmd, self.cmd_sasl)

        cmd = CommandParser(
            prog="CERTFP",
            description="configure CertFP authentication for this network",
            epilog=(
                "Using the set command requires you to paste a bundled PEM certificate (cert + key) on the next line"
                " after the command within the same message. The certificate needs to include both the certificate and"
                " the private key for it to be accepted.\n"
                "\n"
                "Some networks (OFTC) may require you to disable the SASL mechanism, see 'SASL -h' how to change it manually.\n"
                "\n"
                "OpenSSL generation example (from Libera.Chat guides):\n"
                "$ openssl req -x509 -new -newkey rsa:4096 -sha256 -days 1096 -nodes -out libera.pem -keyout libera.pem"
            ),
        )
        cmd.add_argument("--set", action="store_true", help="set X509 certificate bundle (PEM)")
        cmd.add_argument("--remove", action="store_true", help="remove stored certificate")
        self.commands.register(cmd, self.cmd_certfp)

        cmd = CommandParser(
            prog="AUTOCMD",
            description="run commands on connect",
            epilog=(
                "If the network you are connecting to does not support server password to identify you automatically"
                " can set this to send a command before joining channels.\n"
                "If your password contains special characters, use shell escaping rules to escape the entire NICKSERV"
                " command, and its argument within (see example).\n"
                "\n"
                'Example (QuakeNet): AUTOCMD "UMODE +x; MSG -s Q@CServe.quakenet.org auth foo bar"\n'
                "Example (OFTC): AUTOCMD NICKSERV -s identify foo bar\n"
                'Example (special characters): AUTOCMD "NICKSERV -s \\"identify special;\'chars\\""\n'
            ),
        )
        cmd.add_argument("command", nargs="*", help="commands separated with ';'")
        cmd.add_argument("--remove", action="store_true", help="remove stored command")
        self.commands.register(cmd, self.cmd_autocmd)

        cmd = CommandParser(
            prog="CONNECT",
            description="connect to network",
            epilog=(
                "When this command is invoked the connection to this network will be persisted across disconnects and"
                " bridge restart.\n"
                "Only if the server KILLs your connection it will stay disconnected until CONNECT is invoked again.\n"
                "\n"
                "If you want to cancel automatic reconnect you need to issue the DISCONNECT command.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_connect)

        cmd = CommandParser(
            prog="DISCONNECT",
            description="disconnect from network",
            epilog=(
                "In addition to disconnecting from an active network connection this will also cancel any automatic"
                "reconnection attempt.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_disconnect)

        cmd = CommandParser(prog="RECONNECT", description="reconnect to network")
        self.commands.register(cmd, self.cmd_reconnect)

        cmd = CommandParser(
            prog="RAW",
            description="send raw IRC commands",
            epilog=(
                "Arguments (text) are not quoted in any way so it's possible to send ANY command to the server.\n"
                "This is meant as a last resort if the bridge does not have built-in support for some IRC command.\n"
                "\n"
                "Note: You may need to use colon (:) for multi-word arguments, see the IRC RFC for details.\n"
            ),
        )
        cmd.add_argument("text", nargs="+", help="raw text")
        self.commands.register(cmd, self.cmd_raw)

        cmd = CommandParser(
            prog="QUERY",
            description="start a private chat",
            epilog=(
                "Creates a new DM with the target nick. They do not need to be connected for this command to work.\n"
            ),
        )
        cmd.add_argument("nick", help="target nickname")
        cmd.add_argument("message", nargs="*", help="optional message")
        self.commands.register(cmd, self.cmd_query)

        cmd = CommandParser(
            prog="MSG",
            description="send a message without opening a DM",
            epilog=(
                "If the target nick does not exist on the network an error reply may be generated by the server.\n"
            ),
        )
        cmd.add_argument("-s", "--sensitive", action="store_true", help="hide message content from network room")
        cmd.add_argument("nick", help="target nickname")
        cmd.add_argument("message", nargs="+", help="message")
        self.commands.register(cmd, self.cmd_msg)

        cmd = CommandParser(
            prog="CTCP",
            description="send a CTCP command",
            epilog="You probably know what you are doing.",
        )
        cmd.add_argument("nick", help="target nickname")
        cmd.add_argument("command", nargs="+", help="command and arguments")
        self.commands.register(cmd, self.cmd_ctcp)

        cmd = CommandParser(
            prog="CTCPCFG",
            description="enable/disable automatic CTCP replies",
        )
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable CTCP replies")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable CTCP replies")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_ctcpcfg)

        cmd = CommandParser(
            prog="NICKSERV",
            description="send a message to NickServ (if supported by network)",
            epilog=(
                "If your password contains special characters, use shell escaping.\n"
                'Example: NICKSERV "identify pass;word"\n'
                "Alias: NS"
            ),
        )
        cmd.add_argument("-s", "--sensitive", action="store_true", help="hide message content from network room")
        cmd.add_argument("message", nargs="+", help="message")
        self.commands.register(cmd, self.cmd_nickserv, ["NS"])

        cmd = CommandParser(
            prog="CHANSERV",
            description="send a message to ChanServ (if supported by network)",
            epilog="Alias: CS",
        )
        cmd.add_argument("message", nargs="+", help="message")
        self.commands.register(cmd, self.cmd_chanserv, ["CS"])

        cmd = CommandParser(
            prog="JOIN",
            description="join a channel",
            epilog=(
                "Any channels joined will be persisted between reconnects.\n"
                "\n"
                "Note: Bridge administrators can trivially see the stored channel key if they want to.\n"
            ),
        )
        cmd.add_argument("channel", help="target channel")
        cmd.add_argument("key", nargs="?", help="channel key")
        self.commands.register(cmd, self.cmd_join)

        cmd = CommandParser(
            prog="PLUMB",
            description="plumb a room",
            epilog=(
                "Plumbs a channel in single-puppeted mode. This will make the bridge join the room and then join the"
                " configured IRC channel.\n"
            ),
        )
        cmd.add_argument("room", help="target Matrix room ID (eg. !uniqueid:your-homeserver)")
        cmd.add_argument("channel", help="target channel")
        cmd.add_argument("key", nargs="?", help="channel key")
        self.commands.register(cmd, self.cmd_plumb)

        cmd = CommandParser(
            prog="UNPLUMB",
            description="unplumb a room",
        )
        cmd.add_argument("channel", help="target channel")
        self.commands.register(cmd, self.cmd_unplumb)

        cmd = CommandParser(prog="UMODE", description="set user modes")
        cmd.add_argument("flags", help="user mode flags")
        self.commands.register(cmd, self.cmd_umode)

        cmd = CommandParser(
            prog="WAIT",
            description="wait specified amount of time",
            epilog=("Use with AUTOCMD to add delays between commands."),
        )
        cmd.add_argument("seconds", help="how many seconds to wait")
        self.commands.register(cmd, self.cmd_wait)

        cmd = CommandParser(
            prog="PILLS",
            description="configure automatic pills",
        )
        cmd.add_argument(
            "--length", help="minimum length of nick to generate a pill, setting to 0 disables this feature", type=int
        )
        cmd.add_argument("--ignore", help="comma separated list of nicks to ignore for pills")
        self.commands.register(cmd, self.cmd_pills)

        cmd = CommandParser(
            prog="AUTOQUERY",
            description="enable or disable automatic room creation when getting a message",
        )
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable autoquery")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable autoquery")
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_autoquery)

        cmd = CommandParser(prog="WHOIS", description="send a WHOIS(IS) command")
        cmd.add_argument("nick", help="target nick")
        self.commands.register(cmd, self.cmd_whois)

        cmd = CommandParser(prog="WHOAMI", description="send a WHOIS(IS) for ourself")
        self.commands.register(cmd, self.cmd_whoami)

        cmd = CommandParser(
            prog="ROOM",
            description="run a room command from network room",
            epilog=(
                "Try 'ROOM #foo' to get the list of commands for a room."
                "If a command generates IRC replies in a bouncer room they will appear in the room itself."
            ),
        )
        cmd.add_argument("target", help="IRC channel or nick that has a room")
        cmd.add_argument("command", help="Command and arguments", nargs=argparse.REMAINDER)
        self.commands.register(cmd, self.cmd_room)

        cmd = CommandParser(
            prog="AVATAR",
            description="change or show IRC network ghost avatar",
            epilog="Note: This changes the avatar for everyone using this bridge, use with caution.",
        )
        cmd.add_argument("nick", help="nick")
        cmd.add_argument("url", nargs="?", help="new avatar URL (mxc:// format)")
        cmd.add_argument("--remove", help="remove avatar", action="store_true")
        self.commands.register(cmd, self.cmd_avatar)

        cmd = CommandParser(prog="REJOIN", description="configure rejoin behavior for channel rooms")
        cmd.add_argument("--enable-invite", dest="invite", action="store_true", help="Enable rejoin on invite")
        cmd.add_argument("--disable-invite", dest="invite", action="store_false", help="Disable rejoin on invite")
        cmd.add_argument("--enable-kick", dest="kick", action="store_true", help="Enable rejoin on kick")
        cmd.add_argument("--disable-kick", dest="kick", action="store_false", help="Disable rejoin on kick")
        cmd.set_defaults(invite=None, kick=None)
        self.commands.register(cmd, self.cmd_rejoin)

        cmd = CommandParser(prog="STATUS", description="show current network status")
        self.commands.register(cmd, self.cmd_status)

        cmd = CommandParser(prog="SPACE", description="create a managed Matrix space for this network")
        self.commands.register(cmd, self.cmd_space)

        cmd = CommandParser(
            prog="CAPS",
            description="request server capabilities on connect",
            epilog="Only bridge supported capabilities can be requested.",
        )
        cmd.add_argument(
            "--add",
            nargs=1,
            choices=[
                "message-tags",
                "chghost",
                "draft/relaymsg",
                "znc.in/self-message",
                "twitch.tv/membership",
                "away-notify",
            ],
            help="Add to CAP request",
        )
        cmd.add_argument("--remove", nargs=1, help="Remove from CAP request")
        cmd.set_defaults(add=None, remove=None)
        self.commands.register(cmd, self.cmd_caps)

        cmd = CommandParser(
            prog="FORWARD",
            description="configure channel IRC events forwarding",
            epilog="Enabling this will move all IRC noise from channel rooms into network room.",
        )
        cmd.add_argument("--enable", dest="forward", action="store_true", help="Enable forwarding")
        cmd.add_argument("--disable", dest="forward", action="store_false", help="Disable forwarding")
        cmd.set_defaults(forward=None)
        self.commands.register(cmd, self.cmd_forward)

        cmd = CommandParser(
            prog="COLOR",
            description="enable or disable color formatting from IRC",
        )
        cmd.add_argument("--enable", dest="enabled", action="store_true", help="Enable color")
        cmd.add_argument("--disable", dest="enabled", action="store_false", help="Disable color")
        cmd.set_defaults(enabled=True)
        self.commands.register(cmd, self.cmd_color)

        self.mx_register("m.room.message", self.on_mx_message)

    @staticmethod
    async def create(serv, network, user_id, name):
        room_id = await serv.create_room(name, "Network room for {}".format(network), [user_id])
        room = NetworkRoom(room_id, user_id, serv, [serv.user_id, user_id], bans=[])
        room.from_config({"name": network})
        await room.save()
        serv.register_room(room)
        await room.show_help()
        return room

    def from_config(self, config: dict):
        if "name" in config:
            self.name = config["name"]
        else:
            raise Exception("No name key in config for NetworkRoom")

        if "connected" in config:
            self.connected = config["connected"]

        if "nick" in config:
            self.nick = config["nick"]

        if "username" in config:
            self.username = config["username"]

        if "ircname" in config:
            self.ircname = config["ircname"]

        if "password" in config:
            self.password = config["password"]

        if "sasl_mechanism" in config:
            self.sasl_mechanism = config["sasl_mechanism"]

        if "sasl_username" in config:
            self.sasl_username = config["sasl_username"]

        if "sasl_password" in config:
            self.sasl_password = config["sasl_password"]

        if "autocmd" in config:
            self.autocmd = config["autocmd"]

        if "pills_length" in config:
            self.pills_length = config["pills_length"]

        if "pills_ignore" in config:
            self.pills_ignore = config["pills_ignore"]

        if "autoquery" in config:
            self.autoquery = config["autoquery"]

        if "color" in config:
            self.autoquery = config["color"]

        if "allow_ctcp" in config:
            self.allow_ctcp = config["allow_ctcp"]

        if "tls_cert" in config:
            self.tls_cert = config["tls_cert"]

        if "rejoin_invite" in config:
            self.rejoin_invite = config["rejoin_invite"]

        if "rejoin_kick" in config:
            self.rejoin_kick = config["rejoin_kick"]

        if "caps" in config:
            self.caps = config["caps"]

        if "forward" in config:
            self.forward = config["forward"]

    def to_config(self) -> dict:
        return {
            "name": self.name,
            "connected": self.connected,
            "nick": self.nick,
            "username": self.username,
            "ircname": self.ircname,
            "password": self.password,
            "sasl_mechanism": self.sasl_mechanism,
            "sasl_username": self.sasl_username,
            "sasl_password": self.sasl_password,
            "autocmd": self.autocmd,
            "allow_ctcp": self.allow_ctcp,
            "tls_cert": self.tls_cert,
            "pills_length": self.pills_length,
            "pills_ignore": self.pills_ignore,
            "autoquery": self.autoquery,
            "color": self.color,
            "rejoin_invite": self.rejoin_invite,
            "rejoin_kick": self.rejoin_kick,
            "caps": self.caps,
            "forward": self.forward,
        }

    def is_valid(self) -> bool:
        if self.name is None:
            return False

        # we require user to be in network room or be connected with channels or PMs
        if not self.in_room(self.user_id):

            # if not connected (or trying to) we can clean up
            if not self.connected:
                return False

            # only if post_init has been done and we're connected with no rooms clean up
            if self.post_init_done and self.connected and len(self.rooms) == 0:
                return False

        return True

    def cleanup(self) -> None:
        logging.debug(f"Network {self.id} cleaning up")

        # prevent reconnecting ever again
        self.connected = False
        self.disconnect = True

        if self.caps_task:
            self.caps_task.cancel()
            self.caps_task = None
            logging.debug("... cancelled caps task")

        if self.backoff_task:
            self.backoff_task.cancel()
            self.backoff_task = None
            logging.debug("... cancelled backoff task")

        if self.conn:
            self.conn.disconnect()
            logging.debug("... disconnected from IRC network")

        if self.space:
            self.serv.unregister_room(self.space.id)
            self.space.cleanup()
            asyncio.ensure_future(self.serv.leave_room(self.space.id, self.space.members))
            logging.debug("... cleaned up space")

        super().cleanup()

    async def show_help(self):
        self.send_notice_html(f"Welcome to the network room for <b>{html.escape(self.name)}</b>!")

        try:
            return await self.commands.trigger("HELP")
        except CommandParserError as e:
            return self.send_notice(str(e))

    async def on_mx_message(self, event) -> None:
        if str(event.content.msgtype) != "m.text" or event.sender == self.serv.user_id:
            return

        # ignore edits
        if event.content.get_edit():
            return

        try:
            if event.content.formatted_body:
                lines = str(await self.parser.parse(event.content.formatted_body)).split("\n")
            else:
                lines = event.content.body.split("\n")

            command = lines.pop(0)
            tail = "\n".join(lines) if len(lines) > 0 else None

            await self.commands.trigger(command, tail)
        except CommandParserError as e:
            self.send_notice(str(e))

    async def cmd_connect(self, args) -> None:
        await self.connect()

    async def cmd_disconnect(self, args) -> None:
        self.disconnect = True

        if self.backoff_task:
            self.backoff_task.cancel()
            self.backoff_task = None

        if self.caps_task:
            self.caps_task.cancel()
            self.caps_task = None

        self.backoff = 0
        self.next_server = 0
        self.connected_at = 0

        if self.connected:
            self.connected = False
            await self.save()

        if self.conn:
            self.send_notice("Disconnecting...")
            self.conn.disconnect()

    @connected
    async def cmd_reconnect(self, args) -> None:
        await self.cmd_disconnect(Namespace())
        await self.cmd_connect(Namespace())

    @connected
    async def cmd_raw(self, args) -> None:
        self.conn.send_raw(" ".join(args.text))

    @connected
    async def cmd_query(self, args) -> None:
        # TODO: validate nick doesn't look like a channel
        target = args.nick.lower()
        message = " ".join(args.message)

        if target in self.rooms:
            room = self.rooms[target]
            await self.az.intent.invite_user(room.id, self.user_id)
            self.send_notice("Inviting back to private chat with {}.".format(args.nick))
        else:
            room = PrivateRoom.create(self, args.nick)
            self.rooms[room.name] = room
            self.send_notice("You have been invited to private chat with {}.".format(args.nick))

        if len(message) > 0:
            self.conn.privmsg(target, message)
            self.send_notice(f"Sent out-of-room message to {target}: {message}")

    @connected
    async def cmd_msg(self, args) -> None:
        message = " ".join(args.message)
        self.conn.privmsg(args.nick, message)

        if args.sensitive:
            message = "***"

        self.send_notice(f"{self.conn.real_nickname} -> {args.nick}: {message}")

    @connected
    async def cmd_ctcp(self, args) -> None:
        command = args.command[0].upper()
        command_args = " ".join(args.command[1:])
        self.conn.ctcp(command, args.nick, command_args)
        self.send_notice_html(
            f"{self.conn.real_nickname} -> <b>{args.nick}</b> CTCP <b>{html.escape(command)}</b> {html.escape(command_args)}"
        )

    async def cmd_ctcpcfg(self, args) -> None:
        if args.enabled is not None:
            self.allow_ctcp = args.enabled
            await self.save()

        self.send_notice(f"CTCP replies are {'enabled' if self.allow_ctcp else 'disabled'}")

    @connected
    async def cmd_nickserv(self, args) -> None:
        message = " ".join(args.message)
        self.conn.send_raw("NICKSERV " + message)

        if args.sensitive:
            message = "***"

        self.send_notice(f"{self.conn.real_nickname} -> NickServ: {message}")

    @connected
    async def cmd_chanserv(self, args) -> None:
        message = " ".join(args.message)
        self.conn.send_raw("CHANSERV " + message)
        self.send_notice(f"{self.conn.real_nickname} -> ChanServ: {message}")

    @connected
    async def cmd_join(self, args) -> None:
        channel = args.channel

        if re.match(r"^[A-Za-z0-9]", channel):
            channel = "#" + channel

        # cache key so we can store later if join succeeds
        self.keys[channel.lower()] = args.key

        self.conn.join(channel, args.key)

    @connected
    async def cmd_plumb(self, args) -> None:
        channel = args.channel

        if re.match(r"^[A-Za-z0-9]", channel):
            channel = "#" + channel

        if not self.serv.is_admin(self.user_id):
            self.send_notice("Plumbing is currently reserved for admins only.")
            return

        if channel.lower() in self.rooms:
            self.send_notice(f"You are already on {channel}")
            return

        room = await PlumbedRoom.create(id=args.room, network=self, channel=channel, key=args.key)
        self.conn.join(room.name, room.key)

    async def cmd_unplumb(self, args) -> None:
        channel = args.channel.lower()

        if channel not in self.rooms or type(self.rooms[channel]) != PlumbedRoom:
            self.send_notice(f"{args.channel} is not plumbed")
            return

        room = self.rooms[channel]

        self.send_notice(f"Leaving {room.id} to unplumb.")
        await self.az.intent.leave_room(room.id)

    @connected
    async def cmd_umode(self, args) -> None:
        self.conn.mode(self.conn.real_nickname, args.flags)

    async def cmd_wait(self, args) -> None:
        try:
            seconds = float(args.seconds)
            if seconds > 0 and seconds < 30:
                await asyncio.sleep(seconds)
            else:
                self.send_notice(f"Unreasonable wait time: {args.seconds}")
        except ValueError:
            self.send_notice(f"Invalid wait time: {args.seconds}")

    def get_nick(self):
        if self.nick:
            return self.nick

        return self.user_id.split(":", 1)[0][1:]

    async def cmd_nick(self, args) -> None:
        if args.nick is None:
            nick = self.get_nick()
            if self.conn and self.conn.connected:
                self.send_notice(f"Current nickname: {self.conn.real_nickname} (configured: {nick})")
            else:
                self.send_notice(f"Configured nickname: {nick}")
            return

        self.nick = args.nick
        await self.save()
        self.send_notice("Nickname set to {}".format(self.nick))

        if self.conn and self.conn.connected:
            if self.keepnick_task:
                self.keepnick_task.cancel()
                self.keepnick_task = None

            self.conn.nick(args.nick)

    def get_ident(self):
        idents = self.serv.config["idents"]

        # use admin set override if exists
        if self.user_id in idents:
            return idents[self.user_id][:8]

        # return mxid digest if no custom ident
        return (
            "m-"
            + b32encode(hashlib.sha1(self.user_id.encode("utf-8")).digest())
            .decode("utf-8")
            .replace("=", "")[:6]
            .lower()
        )

    async def cmd_username(self, args) -> None:
        if args.remove:
            self.username = None
            await self.save()
            self.send_notice("Username removed.")
            return

        if args.username is None:
            self.send_notice(f"Configured username: {str(self.username)}")
            return

        self.username = args.username
        await self.save()
        self.send_notice(f"Username set to {self.username}")

    async def cmd_realname(self, args) -> None:
        if args.remove:
            self.ircname = None
            await self.save()
            self.send_notice("Realname removed.")
            return

        if args.name is None:
            if self.ircname:
                self.send_notice(f"Configured realname: {self.ircname}")
            else:
                self.send_notice("No configured realname.")
            return

        self.ircname = args.name
        await self.save()
        self.send_notice(f"Realname set to {self.ircname}")

    async def cmd_password(self, args) -> None:
        if args.remove:
            self.password = None
            await self.save()
            self.send_notice("Password removed.")
            return

        if args.password is None:
            self.send_notice(f"Configured password: {self.password if self.password else ''}")
            return

        self.password = args.password
        await self.save()
        self.send_notice(f"Password set to {self.password}")

    async def cmd_sasl(self, args) -> None:
        if args.remove:
            self.sasl_username = None
            self.sasl_password = None
            await self.save()
            self.send_notice("SASL credentials removed.")
            return

        if args.mechanism is None and args.username is None and args.password is None:
            self.send_notice(f"SASL mechanism: {self.sasl_mechanism if self.sasl_mechanism else 'auto'}")
            self.send_notice(f"SASL username: {self.sasl_username}")
            self.send_notice(f"SASL password: {self.sasl_password}")
            return

        if args.mechanism:
            self.sasl_mechanism = args.mechanism if args.mechanism != "auto" else None

        if args.username:
            self.sasl_username = args.username

        if args.password:
            self.sasl_password = args.password

        await self.save()
        self.send_notice("SASL credentials updated.")

    async def cmd_certfp(self, args) -> None:
        if args.remove:
            self.tls_cert = None
            await self.save()
            self.send_notice("CertFP certificate removed.")
        elif args.set:
            if args._tail is None:
                example = (
                    "CERTFP --set\n"
                    "-----BEGIN CERTIFICATE-----\n"
                    "...\n"
                    "-----END CERTIFICATE-----\n"
                    "-----BEGIN PRIVATE KEY-----\n"
                    "...\n"
                    "-----END PRIVATE KEY-----\n"
                )
                self.send_notice_html(
                    f"<p>Expected the certificate to follow command. Certificate not updated.</p><pre><code>{example}</code></pre>"
                )
                return

            # simple sanity checks it possibly looks alright
            if not args._tail.startswith("-----"):
                self.send_notice("This does not look like a PEM certificate.")
                return

            if "-----BEGIN CERTIFICATE----" not in args._tail:
                self.send_notice("Certificate section is missing.")
                return

            if "PRIVATE KEY----" not in args._tail:
                self.send_notice("Private key section is missing.")
                return

            self.tls_cert = args._tail
            await self.save()
            self.send_notice("Client certificate saved.")
        else:
            if self.tls_cert:
                self.send_notice("CertFP certificate exists.")
            else:
                self.send_notice("CertFP certificate does not exist.")

    async def cmd_autocmd(self, args) -> None:
        autocmd = " ".join(args.command)

        if args.remove:
            self.autocmd = None
            await self.save()
            self.send_notice("Autocmd removed.")
            return

        if autocmd == "":
            self.send_notice(f"Configured autocmd: {self.autocmd if self.autocmd else ''}")
            return

        self.autocmd = autocmd
        await self.save()
        self.send_notice(f"Autocmd set to {self.autocmd}")

    @connected
    async def cmd_whois(self, args) -> None:
        self.conn.whois(f"{args.nick} {args.nick}")

    @connected
    async def cmd_whoami(self, args) -> None:
        self.conn.whois(f"{self.conn.real_nickname} {self.conn.real_nickname}")

    async def cmd_room(self, args) -> None:
        target = args.target.lower()

        if target not in self.rooms:
            self.send_notice(f"No room for {args.target}")
            return

        room = self.rooms[target]

        if len(args.command) == 0:
            args.command = ["HELP"]

        await room.commands.trigger_args(args.command, forward=True)

    async def cmd_avatar(self, args):
        if not self.serv.is_admin(self.user_id):
            self.send_notice("Setting avatars is reserved for admins only.")
            return

        # ensure the ghost exists
        irc_user_id = await self.serv.ensure_irc_user_id(self.name, args.nick, update_cache=False)

        if args.remove:
            await self.az.intent.user(irc_user_id).set_avatar_url("")
            self.send_notice("Avatar removed.")
        elif args.url:
            await self.az.intent.user(irc_user_id).set_avatar_url(args.url)
            self.send_notice("Avatar updated.")
        else:
            avatar_url = await self.az.intent.user(irc_user_id).get_avatar_url(irc_user_id)
            if avatar_url:
                self.send_notice(f"Current avatar for {args.nick} is {avatar_url}")
            else:
                self.send_notice(f"{args.nick} does not have a custom avatar.")

    async def cmd_pills(self, args) -> None:
        save = False

        if args.length is not None:
            self.pills_length = args.length
            self.send_notice(f"Pills minimum length set to {self.pills_length}")
            save = True
        else:
            self.send_notice(f"Pills minimum length is {self.pills_length}")

        if args.ignore is not None:
            self.pills_ignore = list(map(lambda x: x.strip(), args.ignore.split(",")))
            self.send_notice(f"Pills ignore list set to {', '.join(self.pills_ignore)}")
            save = True
        else:
            if len(self.pills_ignore) == 0:
                self.send_notice("Pills ignore list is empty.")
            else:
                self.send_notice(f"Pills ignore list: {', '.join(self.pills_ignore)}")

        if save:
            await self.save()

    async def cmd_autoquery(self, args) -> None:
        if args.enabled is not None:
            self.autoquery = args.enabled
            await self.save()

        self.send_notice(f"Autoquery is {'enabled' if self.autoquery else 'disabled'}")

    async def cmd_rejoin(self, args) -> None:
        if args.invite is not None:
            self.rejoin_invite = args.invite
            await self.save()

        if args.kick is not None:
            self.rejoin_kick = args.kick
            await self.save()

        self.send_notice(f"Rejoin on invite is {'enabled' if self.rejoin_invite else 'disabled'}")
        self.send_notice(f"Rejoin on kick is {'enabled' if self.rejoin_kick else 'disabled'}")

    async def cmd_status(self, args) -> None:
        if self.connected_at > 0:
            conntime = asyncio.get_running_loop().time() - self.connected_at
            conntime = str(datetime.timedelta(seconds=int(conntime)))
            self.send_notice(f"Connected for {conntime}")

            if self.real_host[0] != "?":
                self.send_notice(f"Connected from host {self.real_host}")
        else:
            self.send_notice("Not connected to server.")

        await self.cmd_nick(Namespace(nick=None))

        pms = []
        chans = []
        plumbs = []

        for room in self.rooms.values():
            if type(room) == PrivateRoom:
                pms.append(room.name)
            elif type(room) == ChannelRoom:
                chans.append(room.name)
            elif type(room) == PlumbedRoom:
                plumbs.append(room.name)

        if len(chans) > 0:
            self.send_notice(f"Channels: {', '.join(chans)}")

        if len(plumbs) > 0:
            self.send_notice(f"Plumbs: {', '.join(plumbs)}")

        if len(pms) > 0:
            self.send_notice(f"PMs: {', '.join(pms)}")

    async def cmd_space(self, args) -> None:
        if self.space is None:
            # sync create to prevent race conditions
            self.space = SpaceRoom.create(
                self, [room.id for room in self.rooms.values() if type(room) is not PlumbedRoom]
            )

            # calls the api and attaches rooms
            self.send_notice("Creating space and inviting you to it.")
            await self.space.create_finalize()
        else:
            self.send_notice(f"Space already exists ({self.space.id}).")

    async def cmd_caps(self, args) -> None:
        if args.add and args.add[0] not in self.caps:
            self.caps += args.add
            await self.save()
        if args.remove and args.remove[0] in self.caps:
            self.caps.remove(args.remove[0])
            await self.save()

        self.send_notice(f"Capabilities to request: {', '.join(self.caps)}")

        if self.conn and self.conn.connected:
            self.send_notice(f"Capabilities supported: {', '.join(self.caps_supported)}")
            self.send_notice(f"Capabilities enabled: {', '.join(self.caps_enabled)}")

    async def cmd_forward(self, args) -> None:
        if args.forward is not None:
            self.forward = args.forward
            await self.save()

        self.send_notice(f"IRC event forwarding is {'enabled' if self.forward else 'disabled'}")

    async def cmd_color(self, args) -> None:
        if args.enabled is not None:
            self.color = args.enabled
            await self.save()

        self.send_notice(f"Color is {'enabled' if self.color else 'disabled'}")

    def kickban(self, channel: str, nick: str, reason: str) -> None:
        self.pending_kickbans[nick].append((channel, reason))
        self.conn.whois(f"{nick}")

    def _do_kickban(self, channel: str, user_data: Dict[str, str], reason: str) -> None:
        self.conn.mode(channel, f"+b *!*@{user_data['host']}")
        self.conn.kick(channel, user_data["nick"], reason)

    async def connect(self) -> None:
        if not self.is_valid():
            logging.warning("Trying to connect an invalid network {self.id}, this is likely a dangling network.")
            return

        if self.connlock.locked():
            self.send_notice("Already connecting.")
            return

        async with self.connlock:
            if self.conn and self.conn.connected:
                self.send_notice("Already connected.")
                return

            self.disconnect = False
            await self._connect()

    async def post_init(self) -> None:
        # attach loose sub-rooms to us
        for type in [PrivateRoom, ChannelRoom, PlumbedRoom]:
            for room in self.serv.find_rooms(type, self.user_id):
                if room.name not in self.rooms and (
                    room.network_id == self.id or (room.network_id is None and room.network_name == self.name)
                ):
                    room.network = self
                    # this doubles as a migration
                    if room.network_id is None:
                        logging.debug(f"{self.id} attaching and migrating {room.id}")
                        room.network_id = self.id
                        await room.save()
                    else:
                        logging.debug(f"{self.id} attaching {room.id}")
                    self.rooms[room.name] = room

        self.post_init_done = True

    async def _connect(self) -> None:
        # force cleanup
        if self.conn:
            self.conn.close()
            self.conn = None

        network = self.serv.config["networks"][self.name]

        # reset whois and kickbans buffers
        self.whois_data.clear()
        self.pending_kickbans.clear()

        while not self.disconnect:
            if self.name not in self.serv.config["networks"]:
                self.send_notice("This network does not exist on this bridge anymore.")
                return

            if len(network["servers"]) == 0:
                self.connected = False
                self.send_notice("No servers to connect for this network.")
                await self.save()
                return

            server = network["servers"][self.next_server % len(network["servers"])]
            self.next_server += 1

            try:
                asyncio.ensure_future(self.serv.push_bridge_state(BridgeStateEvent.CONNECTING, remote_id=self.name))

                with_tls = ""
                ssl_ctx = False
                server_hostname = None
                if server["tls"] or ("tls_insecure" in server and server["tls_insecure"]):
                    ssl_ctx = ssl.create_default_context()
                    if "tls_insecure" in server and server["tls_insecure"]:
                        with_tls = " with insecure TLS"
                        ssl_ctx.check_hostname = False
                        ssl_ctx.verify_mode = ssl.CERT_NONE
                    else:
                        with_tls = " with TLS"
                        ssl_ctx.verify_mode = ssl.CERT_REQUIRED

                    if self.tls_cert:
                        with_tls += " and CertFP"

                        # do this awful hack to allow the SSL stack to load the cert and key
                        cert_file = tempfile.NamedTemporaryFile()
                        cert_file.write(self.tls_cert.encode("utf-8"))
                        cert_file.flush()

                        ssl_ctx.load_cert_chain(cert_file.name)

                        cert_file.close()

                    if "tls_ciphers" in server and server["tls_ciphers"]:
                        with_tls += " using custom cipher list"
                        ssl_ctx.set_ciphers(server["tls_ciphers"])

                    server_hostname = server["address"]

                proxy = None
                sock = None
                address = server["address"]
                port = server["port"]

                with_proxy = ""
                if "proxy" in server and server["proxy"] is not None and len(server["proxy"]) > 0:
                    proxy = Proxy.from_url(server["proxy"])
                    address = port = None
                    with_proxy = " through a SOCKS proxy"

                self.send_notice(f"Connecting to {server['address']}:{server['port']}{with_tls}{with_proxy}...")

                if proxy:
                    sock = await proxy.connect(dest_host=server["address"], dest_port=server["port"])

                sasl_mechanism = None
                if self.sasl_username and self.sasl_password:
                    self.send_notice(f"Using SASL credentials for username {self.sasl_username}")
                    sasl_mechanism = "plain"
                elif self.tls_cert:
                    sasl_mechanism = "external"

                if sasl_mechanism:
                    # if sasl mechanism is overridden, respect that
                    if self.sasl_mechanism == "none":
                        sasl_mechanism = None
                    elif self.sasl_mechanism in ["plain", "external"]:
                        sasl_mechanism = self.sasl_mechanism

                    self.send_notice(f"SASL mechanism set to '{sasl_mechanism if sasl_mechanism else 'none'}'")

                reactor = HeisenReactor(loop=asyncio.get_running_loop())
                irc_server = reactor.server()
                irc_server.buffer_class = buffer.LenientDecodingLineBuffer
                factory = irc.connection.AioFactory(ssl=ssl_ctx, sock=sock, server_hostname=server_hostname)
                self.conn = await irc_server.connect(
                    address,
                    port,
                    self.get_nick(),
                    self.password,
                    username=self.get_ident() if self.username is None else self.username,
                    ircname=self.ircname,
                    connect_factory=factory,
                )

                self.conn.add_global_handler("disconnect", self.on_disconnect)

                self.conn.add_global_handler("welcome", self.on_welcome)
                self.conn.add_global_handler("umodeis", self.on_umodeis)
                self.conn.add_global_handler("channelmodeis", self.on_pass0)
                self.conn.add_global_handler("channelcreate", self.on_pass0)
                self.conn.add_global_handler("notopic", self.on_pass0)
                self.conn.add_global_handler("currenttopic", self.on_pass0)
                self.conn.add_global_handler("topicinfo", self.on_pass0)
                self.conn.add_global_handler("namreply", self.on_pass1)
                self.conn.add_global_handler("endofnames", self.on_pass0)
                self.conn.add_global_handler("banlist", self.on_pass0)
                self.conn.add_global_handler("endofbanlist", self.on_pass0)
                self.conn.add_global_handler("328", self.on_pass0)  # channel URL
                self.conn.add_global_handler("396", self.on_displayed_host)

                # 400-599
                self.conn.add_global_handler("nosuchnick", self.on_pass_if)
                self.conn.add_global_handler("nosuchchannel", self.on_pass_if)
                self.conn.add_global_handler("cannotsendtochan", self.on_pass0)
                self.conn.add_global_handler("nicknameinuse", self.on_nicknameinuse)
                self.conn.add_global_handler("erroneusnickname", self.on_erroneusnickname)
                self.conn.add_global_handler("unavailresource", self.on_unavailresource)
                self.conn.add_global_handler("usernotinchannel", self.on_pass1)
                self.conn.add_global_handler("notonchannel", self.on_pass0)
                self.conn.add_global_handler("useronchannel", self.on_pass1)
                self.conn.add_global_handler("nologin", self.on_pass1)
                self.conn.add_global_handler("keyset", self.on_pass)
                self.conn.add_global_handler("channelisfull", self.on_pass)
                self.conn.add_global_handler("inviteonlychan", self.on_pass)
                self.conn.add_global_handler("bannedfromchan", self.on_pass)
                self.conn.add_global_handler("badchannelkey", self.on_pass0)
                self.conn.add_global_handler("badchanmask", self.on_pass)
                self.conn.add_global_handler("nochanmodes", self.on_pass)
                self.conn.add_global_handler("banlistfull", self.on_pass)
                self.conn.add_global_handler("cannotknock", self.on_pass)
                self.conn.add_global_handler("chanoprivsneeded", self.on_pass0)

                # protocol
                # FIXME: error
                self.conn.add_global_handler("join", self.on_join)
                self.conn.add_global_handler("join", self.on_join_update_host)
                self.conn.add_global_handler("kick", self.on_pass)
                self.conn.add_global_handler("mode", self.on_pass)
                self.conn.add_global_handler("part", self.on_part)
                self.conn.add_global_handler("privmsg", self.on_privmsg)
                self.conn.add_global_handler("privnotice", self.on_privnotice)
                self.conn.add_global_handler("pubmsg", self.on_pass)
                self.conn.add_global_handler("pubnotice", self.on_pass)
                self.conn.add_global_handler("quit", self.on_quit)
                self.conn.add_global_handler("invite", self.on_invite)
                self.conn.add_global_handler("wallops", self.on_wallops)
                self.conn.add_global_handler("chghost", self.on_chghost)
                # FIXME: action
                self.conn.add_global_handler("topic", self.on_pass)
                self.conn.add_global_handler("nick", self.on_nick)
                self.conn.add_global_handler("umode", self.on_umode)

                self.conn.add_global_handler("kill", self.on_kill)
                self.conn.add_global_handler("error", self.on_error)

                # whois
                self.conn.add_global_handler("whoisuser", self.on_whoisuser)
                self.conn.add_global_handler("whoisserver", self.on_whoisserver)
                self.conn.add_global_handler("whoischannels", self.on_whoischannels)
                self.conn.add_global_handler("whoisidle", self.on_whoisidle)
                self.conn.add_global_handler("whoisaccount", self.on_whoisaccount)  # is logged in as
                self.conn.add_global_handler("whoisoperator", self.on_whoisoperator)
                self.conn.add_global_handler("338", self.on_whoisrealhost)  # is actually using host
                self.conn.add_global_handler("away", self.on_away)
                self.conn.add_global_handler("endofwhois", self.on_endofwhois)

                # tags
                self.conn.add_global_handler("tagmsg", self.on_pass_or_ignore)

                # generated
                self.conn.add_global_handler("ctcp", self.on_ctcp)
                self.conn.add_global_handler("ctcpreply", self.on_ctcpreply)
                self.conn.add_global_handler("action", lambda conn, event: None)

                # anything not handled above
                self.conn.add_global_handler("unhandled_events", self.on_server_message)

                if not self.connected:
                    self.connected = True
                    await self.save()

                # awaiting above allows disconnect to happen in-between
                if self.conn is None:
                    return

                self.disconnect = False
                self.connected_at = asyncio.get_running_loop().time()

                # request CAPs
                caps_req = list(self.caps)
                if sasl_mechanism in ["plain", "external"]:
                    caps_req += ["sasl"]

                self.caps_supported = []
                self.caps_enabled = []
                self.caps_task = None
                if caps_req:
                    self.send_notice(f"Capabilities wanted: {', '.join(caps_req)}")

                    try:
                        self.conn.cap("LS")
                        self.caps_task = asyncio.ensure_future(self.conn.expect("cap", 10))
                        (connection, event) = await self.caps_task
                        self.caps_task = None

                        if self.conn is None:
                            return

                        if len(event.arguments) > 1 and event.arguments[0] == "LS":
                            self.caps_supported = event.arguments[1].split()
                            self.send_notice(f"Capabilities supported by server: {', '.join(self.caps_supported)}")

                            # filter all unsupported caps away
                            caps_req = [cap for cap in caps_req if cap in self.caps_supported]

                        if caps_req:
                            self.send_notice(f"Capabilities requested: {', '.join(caps_req)}")
                            self.conn.cap("REQ", *caps_req)

                            self.caps_task = asyncio.ensure_future(self.conn.expect("cap", 10))
                            (connection, event) = await self.caps_task
                            self.caps_task = None

                            if self.conn is None:
                                return

                            if len(event.arguments) > 1:
                                if event.arguments[0] == "ACK":
                                    self.caps_enabled = event.arguments[1].split()
                                    self.send_notice(
                                        f"Capabilities negotiated with server: {', '.join(self.caps_enabled)}"
                                    )
                                elif event.arguments[0] == "NAK":
                                    self.send_notice("Capabilities request was rejected.")
                        else:
                            self.send_notice("No capabilities requested.")
                    except asyncio.CancelledError:
                        logging.debug("Caps request cancelled.")
                        # if we got cancelled just leave quietly
                        return
                    except asyncio.TimeoutError:
                        if self.conn:
                            self.send_notice("Capabilities request timed out, assuming RFC.")
                        else:
                            # if we got disconnected before we timed out just leave
                            logging.debug("Caps request timed out and returned.")
                            return

                    self.conn.cap("END")

                # SASL stuff
                sasl_creds = self.sasl_username is not None and self.sasl_password is not None
                if (sasl_mechanism == "plain" and sasl_creds) or sasl_mechanism == "external":
                    if "sasl" not in self.caps_enabled:
                        raise irc.client.ServerConnectionError("SASL requested but server does not support it.")
                    try:
                        if sasl_mechanism == "plain":
                            self.conn.send_items("AUTHENTICATE PLAIN")
                        else:
                            self.conn.send_items("AUTHENTICATE EXTERNAL")

                        (connection, event) = await self.conn.expect("authenticate")
                        if event.target != "+":
                            raise irc.client.ServerConnectionError("SASL AUTHENTICATE was rejected.")

                        if sasl_mechanism == "plain":
                            sasl = f"{self.sasl_username}\0{self.sasl_username}\0{self.sasl_password}"
                            self.conn.send_items("AUTHENTICATE", base64.b64encode(sasl.encode("utf8")).decode("utf8"))
                        else:
                            self.conn.send_items("AUTHENTICATE", "+")
                        (connection, event) = await self.conn.expect(
                            ["903", "904", "908", "saslsuccess", "saslfail", "saslmechs"]
                        )
                        if event.type not in ["903", "saslsuccess"]:
                            raise irc.client.ServerConnectionError(event.arguments[0])

                    except asyncio.TimeoutError:
                        raise irc.client.ServerConnectionError("SASL authentication timed out.")

                # run connection registration (SASL, user, nick)
                await self.conn.register()

                asyncio.ensure_future(self.serv.push_bridge_state(BridgeStateEvent.CONNECTED, remote_id=self.name))

                return
            except TimeoutError:
                self.send_notice("Connection timed out.")
            except irc.client.ServerConnectionError as e:
                self.send_notice(str(e))
                self.send_notice(f"Failed to connect: {str(e)}")
                # for SASL failures
                if self.conn:
                    if self.conn.connected:
                        self.conn.disconnect()
                    self.conn = None
            except Exception as e:
                self.send_notice(f"Failed to connect: {str(e)}")

            # cancel any pending caps REQ
            if self.caps_task:
                self.caps_task.cancel()
                self.caps_task = None

            if self.backoff < 1800:
                self.backoff += 5

            self.send_notice(f"Trying next server in {self.backoff} seconds...")

            self.backoff_task = asyncio.ensure_future(asyncio.sleep(self.backoff))
            try:
                await self.backoff_task
            except asyncio.CancelledError:
                break
            finally:
                self.backoff_task = None

        self.send_notice("Connection aborted.")

    def on_disconnect(self, conn, event) -> None:
        if self.caps_task:
            self.caps_task.cancel()
            self.caps_task = None

        if self.conn:
            self.conn.disconnect()
            self.conn.close()
            self.conn = None

        # if we were connected for a while, consider the server working
        if self.connected_at > 0 and asyncio.get_running_loop().time() - self.connected_at > 300:
            self.backoff = 0
            self.next_server = 0
            self.connected_at = 0

        if self.connected and not self.disconnect:
            if self.backoff < 1800:
                self.backoff += 5

            self.send_notice(f"Disconnected, reconnecting in {self.backoff} seconds...")

            async def later(self):
                self.backoff_task = asyncio.ensure_future(asyncio.sleep(self.backoff))
                try:
                    await self.backoff_task
                    await self.connect()
                except asyncio.CancelledError:
                    self.send_notice("Reconnect cancelled.")
                finally:
                    self.backoff_task = None

            asyncio.ensure_future(later(self))
            asyncio.ensure_future(
                self.serv.push_bridge_state(BridgeStateEvent.TRANSIENT_DISCONNECT, remote_id=self.name)
            )
        else:
            self.send_notice("Disconnected.")
            asyncio.ensure_future(self.serv.push_bridge_state(BridgeStateEvent.LOGGED_OUT, remote_id=self.name))

    @ircroom_event()
    def on_pass(self, conn, event) -> None:
        logging.warning(f"IRC room event '{event.type}' fell through, target was from command.")
        source = self.source_text(conn, event)
        args = " ".join(event.arguments)
        source = self.source_text(conn, event)
        target = str(event.target)
        self.send_notice_html(f"<b>{source} {event.type} {target}</b> {html.escape(args)}")

    @ircroom_event()
    def on_pass_if(self, conn, event) -> None:
        self.send_notice(" ".join(event.arguments))

    @ircroom_event()
    def on_pass_or_ignore(self, conn, event) -> None:
        pass

    @ircroom_event(target_arg=0)
    def on_pass0(self, conn, event) -> None:
        logging.warning(f"IRC room event '{event.type}' fell through, target was '{event.arguments[0]}'.")
        self.send_notice(" ".join(event.arguments))

    @ircroom_event(target_arg=1)
    def on_pass1(self, conn, event) -> None:
        logging.warning(f"IRC room event '{event.type}' fell through, target was '{event.arguments[1]}'.")
        self.send_notice(" ".join(event.arguments))

    def on_server_message(self, conn, event) -> None:
        # test if the first argument is an ongoing whois target
        if len(event.arguments) > 0 and event.arguments[0].lower() in self.whois_data:
            data = self.whois_data[event.arguments[0].lower()]
            if "extra" not in data:
                data["extra"] = []

            data["extra"].append(" ".join(event.arguments[1:]))
        else:
            self.send_notice(" ".join(event.arguments))

    def on_umodeis(self, conn, event) -> None:
        self.send_notice(f"Your user mode is: {event.arguments[0]}")

    def on_umode(self, conn, event) -> None:
        self.send_notice(f"User mode changed for {event.target}: {event.arguments[0]}")

    def on_displayed_host(self, conn, event) -> None:
        self.send_notice(" ".join(event.arguments))
        if event.target == conn.real_nickname:
            self.real_host = event.arguments[0]

    def source_text(self, conn, event) -> str:
        source = None

        if event.source is not None:
            source = str(event.source.nick)

            if event.source.user is not None and event.source.host is not None:
                source += f" ({event.source.user}@{event.source.host})"
        else:
            source = conn.server

        return source

    @ircroom_event()
    def on_privnotice(self, conn, event) -> None:
        # show unhandled notices in server room
        source = self.source_text(conn, event)
        plain, formatted = parse_irc_formatting(event.arguments[0])
        self.send_notice_html(f"Notice from <b>{source}:</b> {formatted if formatted else html.escape(plain)}")

    @ircroom_event()
    def on_ctcp(self, conn, event) -> None:
        source = self.source_text(conn, event)

        reply = None
        if self.allow_ctcp:
            if event.arguments[0] == "VERSION":
                reply = f"VERSION Heisenbridge v{__version__}"
            elif event.arguments[0] == "PING" and len(event.arguments) > 1:
                reply = f"PING {event.arguments[1]}"
            elif event.arguments[0] == "TIME":
                reply = f"TIME {unix_to_local(time())}"
            else:
                self.send_notice_html(
                    f"<b>{source}</b> requested unknown <b>CTCP {html.escape(' '.join(event.arguments))}</b>"
                )

        if reply is not None:
            self.conn.ctcp_reply(event.source.nick, reply)
            self.send_notice_html(
                f"<b>{source}</b> requested CTCP <b>{html.escape(event.arguments[0])}</b> -> {html.escape(reply)}"
            )
        else:
            if event.arguments[0] == "ACTION":
                # slightly backwards
                target = event.source.nick.lower()

                if target not in self.rooms:

                    if self.autoquery:

                        async def later():
                            # reuse query command to create a room
                            await self.cmd_query(Namespace(nick=event.source.nick, message=[]))

                            # push the message
                            room = self.rooms[target]
                            await self.serv.ensure_irc_user_id(self.name, event.source.nick)
                            room.on_ctcp(conn, event)

                        asyncio.ensure_future(later())
                    else:
                        source = self.source_text(conn, event)
                        self.send_notice_html(
                            f"Emote from <b>{source}:</b> {html.escape(event.source.nick)} {html.escape(event.arguments[1])}"
                        )
                else:
                    room = self.rooms[target]
                    if not room.in_room(self.user_id):
                        asyncio.ensure_future(self.az.intent.invite_user(self.rooms[target].id, self.user_id))
            else:
                self.send_notice_html(
                    f"<b>{source}</b> requested CTCP <b>{html.escape(event.arguments[0])}</b> (ignored)"
                )

    @ircroom_event()
    def on_ctcpreply(self, conn, event) -> None:
        command = event.arguments[0].upper()
        reply = event.arguments[1]

        self.send_notice_html(
            f"CTCP <b>{html.escape(command)}</b> reply from <b>{event.source.nick}</b>: {html.escape(reply)}"
        )

    def on_welcome(self, conn, event) -> None:
        self.on_server_message(conn, event)

        async def later():
            await asyncio.sleep(2)

            if self.autocmd is not None:
                self.send_notice("Executing autocmd and waiting a bit before joining channels...")
                try:
                    await self.commands.trigger(
                        self.autocmd, allowed=["RAW", "MSG", "NICKSERV", "NS", "CHANSERV", "CS", "UMODE", "WAIT"]
                    )
                except Exception as e:
                    self.send_notice(f"Autocmd failed: {str(e)}")
                await asyncio.sleep(4)

            # detect disconnect before we get to join
            if not self.conn or not self.conn.connected:
                return

            channels = []
            keyed_channels = []

            for room in self.rooms.values():
                if type(room) is ChannelRoom or type(room) is PlumbedRoom:
                    if room.key:
                        keyed_channels.append((room.name, room.key))
                    else:
                        channels.append(room.name)

            if len(channels) > 0:
                self.send_notice(f"Joining channels {', '.join(channels)}")
                self.conn.join(",".join(channels))

            if len(keyed_channels) > 0:
                for channel, key in keyed_channels:
                    self.send_notice(f"Joining {channel} with a key")
                    self.conn.join(channel, key)

        asyncio.ensure_future(later())

    @ircroom_event()
    def on_privmsg(self, conn, event) -> None:
        # slightly backwards
        target = event.source.nick.lower()

        if target not in self.rooms:

            if self.autoquery:

                async def later():
                    # reuse query command to create a room
                    await self.cmd_query(Namespace(nick=event.source.nick, message=[]))

                    # push the message
                    room = self.rooms[target]
                    await self.serv.ensure_irc_user_id(self.name, event.source.nick)
                    room.on_privmsg(conn, event)

                asyncio.ensure_future(later())
            else:
                source = self.source_text(conn, event)
                self.send_notice_html(f"Message from <b>{source}:</b> {html.escape(event.arguments[0])}")
        else:
            room = self.rooms[target]
            if not room.in_room(self.user_id):
                asyncio.ensure_future(self.az.intent.invite_user(self.rooms[target].id, self.user_id))

    @ircroom_event()
    def on_join(self, conn, event) -> None:
        target = event.target.lower()

        logging.debug(f"Handling JOIN to {target} by {event.source.nick} (we are {self.conn.real_nickname})")

        # create a ChannelRoom in response to JOIN
        if event.source.nick == self.conn.real_nickname and target not in self.rooms:
            logging.debug("Pre-flight check for JOIN ok, going to create it...")
            self.rooms[target] = ChannelRoom.create(self, event.target)

            # pass this event through
            self.rooms[target].on_join(conn, event)

    def on_join_update_host(self, conn, event) -> None:
        # update for split long
        if event.source.nick == self.conn.real_nickname and (
            self.real_host != event.source.host or self.real_user != event.source.user
        ):
            self.real_host = event.source.host
            self.real_user = event.source.user
            logging.debug(f"Self host updated to '{self.real_host}', user to '{self.real_user}'")

    @ircroom_event()
    def on_part(self, conn, event) -> None:
        if conn.real_nickname == event.source.nick:
            self.send_notice_html(f"You left <b>{html.escape(event.target)}</b>")
        else:
            # should usually never end up here
            self.send_notice_html(f"<b>{html.escape(event.source.nick)}</b> left <b>{html.escape(event.target)}</b>")

    def on_quit(self, conn, event) -> None:
        # leave channels
        for room in self.rooms.values():
            if type(room) is ChannelRoom or type(room) is PlumbedRoom:
                room.on_quit(conn, event)

    def on_nick(self, conn, event) -> None:
        # the IRC library changes real_nickname before running handlers
        if event.target == self.conn.real_nickname:
            logging.debug(f"Detected own nick change to {event.target}")
            if event.target == self.get_nick():
                self.send_notice(f"You're now known as {event.target}")

        old_irc_user_id = self.serv.irc_user_id(self.name, event.source.nick)
        new_irc_user_id = self.serv.irc_user_id(self.name, event.target)

        # special case where only cases change, ensure will update displayname sometime in the future
        if old_irc_user_id == new_irc_user_id:
            asyncio.ensure_future(self.serv.ensure_irc_user_id(self.name, event.target))

        # leave and join channels
        for room in self.rooms.values():
            if type(room) is ChannelRoom or type(room) is PlumbedRoom:
                room.rename(event.source.nick, event.target)

    def on_nicknameinuse(self, conn, event) -> None:
        self.send_notice(f"Nickname {event.arguments[0]} is in use")
        if self.conn.real_nickname == "":
            newnick = event.arguments[0] + "_"
            self.conn.nick(newnick)
        self.keepnick()

    def on_erroneusnickname(self, conn, event) -> None:
        self.send_notice(f"Nickname {event.arguments[0]} is erroneus and was rejected by the server")

    @ircroom_event()
    def on_unavailresource(self, conn, event) -> None:
        if event.arguments[0][0] not in ["#", "!", "&"]:
            self.send_notice(f"Nickname {event.arguments[0]} is currently unavailable")
            if self.conn.real_nickname == "":
                newnick = event.arguments[0] + "_"
                self.conn.nick(newnick)
            self.keepnick()
        else:
            self.send_notice(f"Channel {event.arguments[0]} is currently unavailable")

    def keepnick(self):
        if self.keepnick_task:
            self.keepnick_task.cancel()

        self.send_notice(f"Trying to set nickname to {self.get_nick()} again after five minutes.")

        def try_keepnick():
            self.keepnick_task = None

            if not self.conn or not self.conn.connected:
                return

            self.conn.nick(self.get_nick())

        self.keepnick_task = asyncio.get_running_loop().call_later(300, try_keepnick)

    def on_invite(self, conn, event) -> None:
        rejoin = ""

        target = event.arguments[0].lower()
        if self.rejoin_invite and target in self.rooms:
            self.conn.join(event.arguments[0])
            rejoin = " (rejoin on invite is enabled, joining back)"

        self.send_notice_html(
            f"<b>{event.source.nick}</b> has invited you to <b>{html.escape(event.arguments[0])}</b>{rejoin}"
        )

    def on_wallops(self, conn, event) -> None:
        plain, formatted = parse_irc_formatting(event.target)
        self.send_notice_html(f"<b>WALLOPS {event.source.nick}</b>: {formatted if formatted else html.escape(plain)}")

    def on_chghost(self, conn, event) -> None:
        # update for split long, ignored for all other users as we don't track their username or host
        # we also allow overriding our own hostname before 001 when our real_nickname is empty
        if (
            (event.source.nick == self.conn.real_nickname or self.conn.real_nickname == "")
            and event.arguments
            and (self.real_host != event.arguments[0] or self.real_user != event.target)
        ):
            self.real_host = event.arguments[0]
            self.real_user = event.target
            logging.debug(f"Self host updated to '{self.real_host}', user to '{self.real_user}'")

    @ircroom_event()
    def on_kill(self, conn, event) -> None:
        if event.target == conn.real_nickname:
            source = self.source_text(conn, event)
            self.send_notice_html(f"Killed by <b>{source}</b>: {html.escape(event.arguments[0])}")

            # do not reconnect after KILL
            self.disconnect = True

    def on_error(self, conn, event) -> None:
        self.send_notice_html(f"<b>ERROR</b>: {html.escape(event.target)}")

    def on_whoisuser(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["nick"] = event.arguments[0]
        data["user"] = event.arguments[1]
        data["host"] = event.arguments[2]
        data["realname"] = event.arguments[4]

    def on_whoisserver(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["server"] = f"{event.arguments[1]} ({event.arguments[2]})"

    def on_whoischannels(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["channels"] = event.arguments[1]

    def on_whoisidle(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["idle"] = str(datetime.timedelta(seconds=int(event.arguments[1])))
        if len(event.arguments) > 2:
            data["signon"] = unix_to_local(int(event.arguments[2]))

    def on_whoisaccount(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["account"] = event.arguments[1]

    def on_whoisoperator(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["ircop"] = event.arguments[1]

    def on_whoisrealhost(self, conn, event) -> None:
        data = self.whois_data[event.arguments[0].lower()]
        data["realhost"] = event.arguments[1]

    def on_away(self, conn, event) -> None:
        if len(event.arguments) < 2:
            target = self
            if event.source.nick in self.rooms:
                target = self.rooms[event.source.nick]

            if event.target is None:
                target.send_notice(f"{event.source.nick} is no longer away")
            else:
                target.send_notice(f"{event.source.nick} is now away: {event.target}")
        elif event.arguments[0].lower() in self.whois_data:
            self.whois_data[event.arguments[0].lower()]["away"] = event.arguments[1]
        else:
            self.send_notice(f"{event.arguments[0]} is away: {event.arguments[1]}")

    def on_endofwhois(self, conn, event) -> None:
        nick = event.arguments[0].lower()
        data = self.whois_data[nick]
        del self.whois_data[nick]

        if nick in self.pending_kickbans:
            channels = self.pending_kickbans[nick]
            del self.pending_kickbans[nick]
            for channel, reason in channels:
                self._do_kickban(channel, data, reason)
            return

        reply = []
        fallback = []
        reply.append("<table>")

        for k in [
            "nick",
            "user",
            "host",
            "realname",
            "realhost",
            "away",
            "channels",
            "server",
            "ircop",
            "idle",
            "signon",
            "account",
        ]:
            if k in data:
                reply.append(f"<tr><td>{k}</td><td>{html.escape(data[k])}</td>")
                fallback.append(f"{k}: {data[k]}")

        if "extra" in data:
            for v in data["extra"]:
                reply.append(f"<tr><td></td><td>{html.escape(v)}</td>")
                fallback.append(f"{data['nick']} {v}")

        reply.append("</table>")

        # forward whois reply to a DM if exists
        target = self
        if nick in self.rooms:
            target = self.rooms[nick]

        target.send_notice(formatted="".join(reply), text="\n".join(fallback))



import re
from typing import Dict
from typing import Optional
from typing import Pattern

from mautrix.types import UserID
from mautrix.util.formatter.formatted_string import EntityType
from mautrix.util.formatter.html_reader import HTMLNode
from mautrix.util.formatter.markdown_string import MarkdownString
from mautrix.util.formatter.parser import MatrixParser
from mautrix.util.formatter.parser import RecursionContext
from mautrix.util.formatter.parser import T


class IRCString(MarkdownString):
    def format(self, entity_type: EntityType, **kwargs) -> "IRCString":
        if entity_type == EntityType.BOLD:
            self.text = f"*{self.text}*"
        elif entity_type == EntityType.ITALIC:
            self.text = f"_{self.text}_"
        elif entity_type == EntityType.STRIKETHROUGH:
            self.text = f"~{self.text}~"
        elif entity_type == EntityType.UNDERLINE:
            self.text = self.text
        elif entity_type == EntityType.URL:
            if kwargs["url"] != self.text:
                self.text = f"{self.text} ({kwargs['url']})"
        elif entity_type == EntityType.EMAIL:
            self.text = self.text
        elif entity_type == EntityType.PREFORMATTED:
            self.text = re.sub(r"\n+", "\n", self.text) + "\n"
        elif entity_type == EntityType.INLINE_CODE:
            self.text = f'"{self.text}"'
        elif entity_type == EntityType.BLOCKQUOTE:
            children = self.trim().split("\n")
            children = [child.prepend("> ") for child in children]
            self.text = self.join(children, "\n").text
        elif entity_type == EntityType.USER_MENTION:
            if kwargs["displayname"] is not None:
                self.text = kwargs["displayname"]

        return self


class IRCMatrixParser(MatrixParser):
    fs = IRCString
    list_bullets = ("-", "*", "+", "=")
    displaynames = Dict[str, str]

    # use .* to account for legacy empty mxid
    mention_regex: Pattern = re.compile("https://matrix.to/#/(@.*:.+)")

    def __init__(self, displaynames: Dict[str, str]) -> T:
        self.displaynames = displaynames

    async def tag_aware_parse_node(self, node: HTMLNode, ctx: RecursionContext) -> T:
        msgs = await self.node_to_tagged_fstrings(node, ctx)
        output = self.fs()
        prev_was_block = True
        for msg, tag in msgs:
            if tag in self.block_tags:
                msg = msg.trim()
                if not prev_was_block:
                    output.append("\n")
                prev_was_block = True
            else:
                prev_was_block = False
            output = output.append(msg)
        return output.trim()

    async def user_pill_to_fstring(self, msg: T, user_id: UserID) -> Optional[T]:
        displayname = None
        if user_id in self.displaynames:
            displayname = self.displaynames[user_id]
        return msg.format(self.e.USER_MENTION, user_id=user_id, displayname=displayname)



import logging
import re
from abc import ABC
from collections import defaultdict
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

from mautrix.appservice import AppService as MauService
from mautrix.types import Membership
from mautrix.types.event.type import EventType

from heisenbridge.appservice import AppService
from heisenbridge.event_queue import EventQueue
from heisenbridge.parser import IRCMatrixParser


class RoomInvalidError(Exception):
    pass


class Room(ABC):
    az: MauService
    id: str
    user_id: str
    serv: AppService
    members: List[str]
    lazy_members: Optional[Dict[str, str]]
    hidden_room_id: Optional[str]
    bans: List[str]
    displaynames: Dict[str, str]
    parser: IRCMatrixParser

    _mx_handlers: Dict[str, List[Callable[[dict], bool]]]
    _queue: EventQueue

    def __init__(self, id: str, user_id: str, serv: AppService, members: List[str], bans: List[str]):
        self.id = id
        self.user_id = user_id
        self.serv = serv
        self.members = list(members)
        self.bans = list(bans) if bans else []
        self.lazy_members = None
        self.hidden_room_id = None
        self.displaynames = {}
        self.last_messages = defaultdict(str)
        self.parser = IRCMatrixParser(self.displaynames)

        self._mx_handlers = {}
        self._queue = EventQueue(self._flush_events)

        # start event queue
        if self.id:
            self._queue.start()

        # we track room members
        self.mx_register("m.room.member", self._on_mx_room_member)

        self.init()

    @classmethod
    def init_class(cls, az: MauService):
        cls.az = az

    async def post_init(self):
        pass

    def from_config(self, config: dict) -> None:
        pass

    def init(self) -> None:
        pass

    def is_valid(self) -> bool:
        return True

    def cleanup(self):
        self._queue.stop()

    def to_config(self) -> dict:
        return {}

    async def save(self) -> None:
        config = self.to_config()
        config["type"] = type(self).__name__
        config["user_id"] = self.user_id
        await self.az.intent.set_account_data("irc", config, self.id)

    def mx_register(self, type: str, func: Callable[[dict], bool]) -> None:
        if type not in self._mx_handlers:
            self._mx_handlers[type] = []

        self._mx_handlers[type].append(func)

    async def on_mx_event(self, event: dict) -> None:
        handlers = self._mx_handlers.get(str(event.type), [self._on_mx_unhandled_event])

        for handler in handlers:
            await handler(event)

    def in_room(self, user_id):
        return user_id in self.members

    async def on_mx_ban(self, user_id) -> None:
        pass

    async def on_mx_unban(self, user_id) -> None:
        pass

    async def on_mx_leave(self, user_id) -> None:
        pass

    async def _on_mx_unhandled_event(self, event: dict) -> None:
        pass

    async def _on_mx_room_member(self, event: dict) -> None:
        if event.content.membership in [Membership.LEAVE, Membership.BAN] and event.state_key in self.members:
            self.members.remove(event.state_key)
            if event.state_key in self.displaynames:
                del self.displaynames[event.state_key]
            if event.state_key in self.last_messages:
                del self.last_messages[event.state_key]

            if not self.is_valid():
                raise RoomInvalidError(
                    f"Room {self.id} ended up invalid after membership change, returning false from event handler."
                )

        if event.content.membership == Membership.LEAVE:
            if event.state_key in self.bans:
                self.bans.remove(event.state_key)
                await self.on_mx_unban(event.state_key)
            else:
                await self.on_mx_leave(event.state_key)

        if event.content.membership == Membership.BAN:
            if event.state_key not in self.bans:
                self.bans.append(event.state_key)

            await self.on_mx_ban(event.state_key)

        if event.content.membership == Membership.JOIN:
            if event.state_key not in self.members:
                self.members.append(event.state_key)

            if event.content.displayname is not None:
                self.displaynames[event.state_key] = str(event.content.displayname)
            elif event.state_key in self.displaynames:
                del self.displaynames[event.state_key]

    async def _join(self, user_id, nick=None):
        if self.hidden_room_id:
            await self.az.intent.user(user_id).ensure_joined(self.hidden_room_id)

        await self.az.intent.user(user_id).ensure_joined(self.id, ignore_cache=True)

        self.members.append(user_id)
        if nick is not None:
            self.displaynames[user_id] = nick

    async def _flush_events(self, events):
        for event in events:
            try:
                if event["type"] == "_join":
                    if self.lazy_members is not None:
                        self.lazy_members[event["user_id"]] = event["nick"]

                    if event["user_id"] not in self.members:
                        if not event["lazy"]:
                            await self._join(event["user_id"], event["nick"])
                elif event["type"] == "_leave":
                    if self.lazy_members is not None and event["user_id"] in self.lazy_members:
                        del self.lazy_members[event["user_id"]]

                    if event["user_id"] in self.members:
                        if event["reason"] is not None:
                            await self.az.intent.user(event["user_id"]).kick_user(
                                self.id, event["user_id"], event["reason"]
                            )
                        else:
                            await self.az.intent.user(event["user_id"]).leave_room(self.id)
                        if event["user_id"] in self.members:
                            self.members.remove(event["user_id"])
                        if event["user_id"] in self.displaynames:
                            del self.displaynames[event["user_id"]]
                elif event["type"] == "_rename":
                    old_irc_user_id = self.serv.irc_user_id(self.network.name, event["old_nick"])
                    new_irc_user_id = self.serv.irc_user_id(self.network.name, event["new_nick"])

                    # if we are lazy loading and this user has never spoken, update that
                    if (
                        self.lazy_members is not None
                        and old_irc_user_id in self.lazy_members
                        and old_irc_user_id not in self.members
                    ):
                        del self.lazy_members[old_irc_user_id]
                        self.lazy_members[new_irc_user_id] = event["new_nick"]
                        continue

                    # this event is created for all rooms, skip if irrelevant
                    if old_irc_user_id not in self.members:
                        continue

                    # always ensure new irc user id is in lazy list
                    if self.lazy_members is not None:
                        self.lazy_members[new_irc_user_id] = event["new_nick"]

                    # check if we can just update the displayname
                    if old_irc_user_id != new_irc_user_id:
                        # ensure we have the new puppet
                        await self.serv.ensure_irc_user_id(self.network.name, event["new_nick"])

                        # old puppet away
                        await self.az.intent.user(old_irc_user_id).kick_user(
                            self.id, old_irc_user_id, f"Changing nick to {event['new_nick']}"
                        )
                        self.members.remove(old_irc_user_id)
                        if old_irc_user_id in self.displaynames:
                            del self.displaynames[old_irc_user_id]

                        # new puppet in
                        if new_irc_user_id not in self.members:
                            await self._join(new_irc_user_id, event["new_nick"])

                elif event["type"] == "_kick":
                    if event["user_id"] in self.members:
                        await self.az.intent.kick_user(self.id, event["user_id"], event["reason"])
                        self.members.remove(event["user_id"])
                        if event["user_id"] in self.displaynames:
                            del self.displaynames[event["user_id"]]
                elif event["type"] == "_ensure_irc_user_id":
                    await self.serv.ensure_irc_user_id(event["network"], event["nick"])
                elif "state_key" in event:
                    intent = self.az.intent

                    if event["user_id"]:
                        intent = intent.user(event["user_id"])

                    await intent.send_state_event(
                        self.id, EventType.find(event["type"]), state_key=event["state_key"], content=event["content"]
                    )
                else:
                    # invite puppet *now* if we are lazy loading and it should be here
                    if (
                        self.lazy_members is not None
                        and event["user_id"] in self.lazy_members
                        and event["user_id"] not in self.members
                    ):
                        await self.serv.ensure_irc_user_id(self.network.name, self.lazy_members[event["user_id"]])
                        await self._join(event["user_id"], self.lazy_members[event["user_id"]])

                    # if we get an event from unknown user (outside room for some reason) we may have a fallback
                    if event["user_id"] is not None and event["user_id"] not in self.members:
                        if "fallback_html" in event and event["fallback_html"] is not None:
                            fallback_html = event["fallback_html"]
                        else:
                            fallback_html = (
                                f"{event['user_id']} sent {event['type']} but is not in the room, this is a bug."
                            )

                        # create fallback event
                        event["content"] = {
                            "msgtype": "m.notice",
                            "body": re.sub("<[^<]+?>", "", event["fallback_html"]),
                            "format": "org.matrix.custom.html",
                            "formatted_body": fallback_html,
                        }

                        # unpuppet
                        event["user_id"] = None

                    intent = self.az.intent.user(event["user_id"]) if event["user_id"] else self.az.intent
                    type = EventType.find(event["type"])
                    await intent.send_message_event(self.id, type, event["content"])
            except Exception:
                logging.exception("Queued event failed")

    # send message to mx user (may be puppeted)
    def send_message(
        self, text: str, user_id: Optional[str] = None, formatted=None, fallback_html: Optional[str] = None
    ) -> None:
        if formatted:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.text",
                    "format": "org.matrix.custom.html",
                    "body": text,
                    "formatted_body": formatted,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }
        else:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.text",
                    "body": text,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }

        self._queue.enqueue(event)

    # send emote to mx user (may be puppeted)
    def send_emote(self, text: str, user_id: Optional[str] = None, fallback_html: Optional[str] = None) -> None:
        event = {
            "type": "m.room.message",
            "content": {
                "msgtype": "m.emote",
                "body": text,
            },
            "user_id": user_id,
            "fallback_html": fallback_html,
        }

        self._queue.enqueue(event)

    # send notice to mx user (may be puppeted)
    def send_notice(
        self, text: str, user_id: Optional[str] = None, formatted=None, fallback_html: Optional[str] = None
    ) -> None:
        if formatted:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.notice",
                    "format": "org.matrix.custom.html",
                    "body": text,
                    "formatted_body": formatted,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }
        else:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.notice",
                    "body": text,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }

        self._queue.enqueue(event)

    # send notice to mx user (may be puppeted)
    def send_notice_html(self, text: str, user_id: Optional[str] = None) -> None:
        event = {
            "type": "m.room.message",
            "content": {
                "msgtype": "m.notice",
                "body": re.sub("<[^<]+?>", "", text),
                "format": "org.matrix.custom.html",
                "formatted_body": text,
            },
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def react(self, event_id: str, text: str) -> None:
        event = {
            "type": "m.reaction",
            "content": {
                "m.relates_to": {
                    "event_id": event_id,
                    "key": text,
                    "rel_type": "m.annotation",
                }
            },
            "user_id": None,
        }

        self._queue.enqueue(event)

    def set_topic(self, topic: str, user_id: Optional[str] = None) -> None:
        event = {
            "type": "m.room.topic",
            "content": {
                "topic": topic,
            },
            "state_key": "",
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def join(self, user_id: str, nick=None, lazy=False) -> None:
        event = {
            "type": "_join",
            "content": {},
            "user_id": user_id,
            "nick": nick,
            "lazy": lazy,
        }

        self._queue.enqueue(event)

    def leave(self, user_id: str, reason: Optional[str] = None) -> None:
        event = {
            "type": "_leave",
            "content": {},
            "reason": reason,
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def rename(self, old_nick: str, new_nick: str) -> None:
        event = {
            "type": "_rename",
            "content": {},
            "old_nick": old_nick,
            "new_nick": new_nick,
        }

        self._queue.enqueue(event)

    def kick(self, user_id: str, reason: str) -> None:
        event = {
            "type": "_kick",
            "content": {},
            "reason": reason,
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def ensure_irc_user_id(self, network, nick):
        event = {
            "type": "_ensure_irc_user_id",
            "content": {},
            "network": network,
            "nick": nick,
            "user_id": None,
        }

        self._queue.enqueue(event)



import logging

from heisenbridge.appservice import AppService
from heisenbridge.room import Room


class HiddenRoom(Room):
    @staticmethod
    async def create(serv: AppService) -> "HiddenRoom":
        logging.debug("HiddenRoom.create(serv)")
        room_id = await serv.create_room("heisenbridge-hidden-room", "Invite-Sink for Heisenbridge", [])
        room = HiddenRoom(
            room_id,
            None,
            serv,
            [serv.user_id],
            [],
        )
        await room.save()
        serv.register_room(room)
        return room

    def is_valid(self) -> bool:
        # Hidden Room usage has been explicitly disabled by user
        if not self.serv.config.get("use_hidden_room", True):
            return False

        # Server already has a (different) hidden room
        if self.serv.hidden_room and self.serv.hidden_room is not self:
            return False

        return True

    async def post_init(self) -> None:
        # Those can be huge lists, but are entirely unused. Free up some memory.
        self.members = []
        self.displaynames = {}



import asyncio
import json
import logging

import aiohttp
from mautrix.types.event import Event


class AppserviceWebsocket:
    def __init__(self, url, token, callback):
        self.url = url + "/_matrix/client/unstable/fi.mau.as_sync"
        self.headers = {
            "Authorization": f"Bearer {token}",
            "X-Mautrix-Websocket-Version": "3",
        }
        self.callback = callback

    async def start(self):
        asyncio.create_task(self._loop())

    async def _loop(self):
        while True:
            try:
                logging.info(f"Connecting to {self.url}...")

                async with aiohttp.ClientSession(headers=self.headers) as sess:
                    async with sess.ws_connect(self.url) as ws:
                        logging.info("Websocket connected.")

                        async for msg in ws:
                            if msg.type != aiohttp.WSMsgType.TEXT:
                                logging.debug("Unhandled WS message: %s", msg)
                                continue

                            data = msg.json()
                            if data["status"] == "ok" and data["command"] == "transaction":
                                logging.debug(f"Websocket transaction {data['txn_id']}")
                                for event in data["events"]:
                                    try:
                                        await self.callback(Event.deserialize(event))
                                    except Exception as e:
                                        logging.error(e)

                                await ws.send_str(
                                    json.dumps(
                                        {
                                            "command": "response",
                                            "id": data["id"],
                                            "data": {},
                                        }
                                    )
                                )
                            else:
                                logging.warn("Unhandled WS command: %s", data)

                logging.info("Websocket disconnected.")
            except asyncio.CancelledError:
                logging.info("Websocket was cancelled.")
                return
            except Exception as e:
                logging.error(e)

                try:
                    await asyncio.sleep(5)
                except asyncio.CancelledError:
                    return



import argparse
import asyncio
import base64
import grp
import hashlib
import hmac
import logging
import os
import pwd
import random
import re
import string
import sys
import urllib
from fnmatch import fnmatch
from typing import Dict
from typing import List
from typing import Tuple

from aiohttp import web
from mautrix.api import HTTPAPI
from mautrix.api import Method
from mautrix.api import Path
from mautrix.api import SynapseAdminPath
from mautrix.appservice import AppService as MauService
from mautrix.appservice.state_store import ASStateStore
from mautrix.client.state_store.memory import MemoryStateStore
from mautrix.errors import MatrixConnectionError
from mautrix.errors import MatrixRequestError
from mautrix.errors import MForbidden
from mautrix.errors import MUserInUse
from mautrix.types import EventType
from mautrix.types import JoinRule
from mautrix.types import Membership
from mautrix.util.bridge_state import BridgeState
from mautrix.util.bridge_state import BridgeStateEvent
from mautrix.util.config import yaml

from heisenbridge import __version__
from heisenbridge.appservice import AppService
from heisenbridge.channel_room import ChannelRoom
from heisenbridge.control_room import ControlRoom
from heisenbridge.hidden_room import HiddenRoom
from heisenbridge.identd import Identd
from heisenbridge.network_room import NetworkRoom
from heisenbridge.plumbed_room import PlumbedRoom
from heisenbridge.private_room import PrivateRoom
from heisenbridge.room import Room
from heisenbridge.room import RoomInvalidError
from heisenbridge.space_room import SpaceRoom
from heisenbridge.websocket import AppserviceWebsocket


class MemoryBridgeStateStore(ASStateStore, MemoryStateStore):
    def __init__(self) -> None:
        ASStateStore.__init__(self)
        MemoryStateStore.__init__(self)


class BridgeAppService(AppService):
    az: MauService
    _api: HTTPAPI
    _rooms: Dict[str, Room]
    _users: Dict[str, str]

    DEFAULT_MEDIA_PATH = "/_heisenbridge/media/{server}/{media_id}/{checksum}{filename}"

    async def push_bridge_state(
        self,
        state_event: BridgeStateEvent,
        error=None,
        message=None,
        ttl=None,
        remote_id=None,
    ) -> None:
        if "heisenbridge" not in self.registration or "status_endpoint" not in self.registration["heisenbridge"]:
            return

        state = BridgeState(
            state_event=state_event,
            error=error,
            message=message,
            ttl=ttl,
            remote_id=remote_id,
        )

        logging.debug(f"Updating bridge state {state}")

        await state.send(self.registration["heisenbridge"]["status_endpoint"], self.az.as_token, log=logging)

    def register_room(self, room: Room):
        self._rooms[room.id] = room

    def unregister_room(self, room_id):
        if room_id in self._rooms:
            del self._rooms[room_id]

    # this is mostly used by network rooms at init, it's a bit slow
    def find_rooms(self, rtype=None, user_id=None) -> List[Room]:
        ret = []

        if rtype is not None and type(rtype) != str:
            rtype = rtype.__name__

        for room in self._rooms.values():
            if (rtype is None or room.__class__.__name__ == rtype) and (user_id is None or room.user_id == user_id):
                ret.append(room)

        return ret

    def is_admin(self, user_id: str):
        if user_id == self.config["owner"]:
            return True

        for mask, value in self.config["allow"].items():
            if fnmatch(user_id, mask) and value == "admin":
                return True

        return False

    def is_user(self, user_id: str):
        if self.is_admin(user_id):
            return True

        for mask in self.config["allow"].keys():
            if fnmatch(user_id, mask):
                return True

        return False

    def is_local(self, mxid: str):
        return mxid.endswith(":" + self.server_name)

    def strip_nick(self, nick: str) -> Tuple[str, str]:
        m = re.match(r"^([~&@%\+!]?)(.+)$", nick)
        if m:
            return (m.group(2), (m.group(1) if len(m.group(1)) > 0 else None))
        else:
            raise TypeError(f"Input nick is not valid: '{nick}'")

    def split_irc_user_id(self, user_id):
        (name, server) = user_id.split(":", 1)

        network = None
        nick = None

        if server != self.server_name:
            return None, None

        if not name.startswith("@" + self.puppet_prefix):
            return None, None

        network_nick = name[len(self.puppet_prefix) + 1 :]

        m = re.match(r"([^" + self.puppet_separator + r"]+).(.+)$", network_nick)

        if m:
            network = re.sub(r"=([0-9a-z]{2})", lambda m: bytes.fromhex(m.group(1)).decode("utf-8"), m.group(1)).lower()
            nick = re.sub(r"=([0-9a-z]{2})", lambda m: bytes.fromhex(m.group(1)).decode("utf-8"), m.group(2)).lower()

        return network, nick

    def nick_from_irc_user_id(self, network, user_id):
        (name, server) = user_id.split(":", 1)

        if server != self.server_name:
            return None

        prefix = "@" + re.sub(
            r"[^0-9a-z\-\.=\_/]",
            lambda m: "=" + m.group(0).encode("utf-8").hex(),
            f"{self.puppet_prefix}{network}{self.puppet_separator}".lower(),
        )

        if not name.startswith(prefix):
            return None

        nick = name[len(prefix) :]
        nick = re.sub(r"=([0-9a-z]{2})", lambda m: bytes.fromhex(m.group(1)).decode("utf-8"), nick)

        return nick

    def irc_user_id(self, network, nick, at=True, server=True):
        nick, mode = self.strip_nick(nick)

        ret = re.sub(
            r"[^0-9a-z\-\.=\_/]",
            lambda m: "=" + m.group(0).encode("utf-8").hex(),
            f"{self.puppet_prefix}{network}{self.puppet_separator}{nick}".lower(),
        )

        if at:
            ret = "@" + ret

        if server:
            ret += ":" + self.server_name

        return ret

    async def cache_user(self, user_id, displayname):
        # start by caching that the user_id exists without a displayname
        if user_id not in self._users:
            self._users[user_id] = None

        # if the cached displayname is incorrect
        if displayname and self._users[user_id] != displayname:
            try:
                await self.az.intent.user(user_id).set_displayname(displayname)
                self._users[user_id] = displayname
            except MatrixRequestError as e:
                logging.warning(f"Failed to set displayname '{displayname}' for user_id '{user_id}', got '{e}'")

    def is_user_cached(self, user_id, displayname=None):
        return user_id in self._users and (displayname is None or self._users[user_id] == displayname)

    async def ensure_irc_user_id(self, network, nick, update_cache=True):
        user_id = self.irc_user_id(network, nick)

        # if we've seen this user before, we can skip registering
        if not self.is_user_cached(user_id):
            await self.az.intent.user(self.irc_user_id(network, nick)).ensure_registered()

        # always ensure the displayname is up-to-date
        if update_cache:
            await self.cache_user(user_id, nick)

        return user_id

    async def _on_mx_event(self, event):
        if event.room_id and event.room_id in self._rooms:
            try:
                room = self._rooms[event.room_id]
                await room.on_mx_event(event)
            except RoomInvalidError:
                logging.info(f"Event handler for {event.type} threw RoomInvalidError, leaving and cleaning up.")
                self.unregister_room(room.id)
                room.cleanup()

                await self.leave_room(room.id, room.members)
            except Exception:
                logging.exception("Ignoring exception from room handler. This should be fixed.")
        elif (
            str(event.type) == "m.room.member"
            and event.sender != self.user_id
            and event.content.membership == Membership.INVITE
        ):
            # set owner if we have none and the user is from the same HS
            if self.config.get("owner", None) is None and event.sender.endswith(":" + self.server_name):
                logging.info(f"We have an owner now, let us rejoice, {event.sender}!")
                self.config["owner"] = event.sender
                await self.save()

            if not self.is_user(event.sender):
                logging.info(f"Non-whitelisted user {event.sender} tried to invite us, ignoring.")
                return
            else:
                logging.info(f"Got an invite from {event.sender}")

            if not event.content.is_direct:
                logging.debug("Got an invite to non-direct room, ignoring")
                return

            # only respond to invites unknown new rooms
            if event.room_id in self._rooms:
                logging.debug("Got an invite to room we're already in, ignoring")
                return

            # handle invites against puppets
            if event.state_key != self.user_id:
                logging.info(f"Whitelisted user {event.sender} invited {event.state_key}, going to reject.")

                try:
                    await self.az.intent.user(event.state_key).kick_user(
                        event.room_id,
                        event.state_key,
                        "Will invite YOU instead",
                    )
                except Exception:
                    logging.exception("Failed to reject invitation.")

                (network, nick) = self.split_irc_user_id(event.state_key)

                if network is not None and nick is not None:
                    for room in self.find_rooms(NetworkRoom, event.sender):
                        if room.name.lower() == network.lower():
                            logging.debug(
                                "Found matching network room ({network}) for {event.sender}, emulating query command for {nick}"
                            )
                            await room.cmd_query(argparse.Namespace(nick=nick, message=[]))
                            break

                return

            logging.info(f"Whitelisted user {event.sender} invited us, going to accept.")

            # accept invite sequence
            try:
                room = ControlRoom(id=event.room_id, user_id=event.sender, serv=self, members=[event.sender], bans=[])
                await room.save()
                self.register_room(room)

                await self.az.intent.join_room(room.id)

                # show help on open
                await room.show_help()
            except Exception:
                if event.room_id in self._rooms:
                    del self._rooms[event.room_id]
                logging.exception("Failed to create control room.")
        else:
            pass
            # print(json.dumps(event, indent=4, sort_keys=True))

    async def detect_public_endpoint(self):
        async with self.api.session as session:
            # first try https well-known
            try:
                resp = await session.request(
                    "GET",
                    "https://{}/.well-known/matrix/client".format(self.server_name),
                )
                data = await resp.json(content_type=None)
                return data["m.homeserver"]["base_url"]
            except Exception:
                logging.debug("Did not find .well-known for HS")

            # try https directly
            try:
                resp = await session.request("GET", "https://{}/_matrix/client/versions".format(self.server_name))
                await resp.json(content_type=None)
                return "https://{}".format(self.server_name)
            except Exception:
                logging.debug("Could not use direct connection to HS")

            # give up
            logging.warning("Using internal URL for homeserver, media links are likely broken!")
            return str(self.api.base_url)

    def mxc_checksum(self, server: str, media_id: str) -> str:
        checksum_raw = hmac.new(self.media_key, f"mxc://{server}/{media_id}/".encode("utf-8"), hashlib.sha256).digest()
        return base64.urlsafe_b64encode(checksum_raw[:8]).decode("utf-8").rstrip("=")

    async def proxy_media(self, req: web.Request) -> web.StreamResponse | web.Response:
        server = req.match_info["server"]
        media_id = req.match_info["media_id"]
        checksum = req.match_info["checksum"]
        if self.mxc_checksum(server, media_id) != checksum:
            return web.Response(status=403, text="Invalid checksum")
        download_url = self.api.base_url / "_matrix/client/v1/media/download" / server / media_id
        filename = req.match_info.get("filename", "")
        if filename:
            download_url /= filename
        query_params: dict[str, str] = {"allow_redirect": "true", "user_id": self.az.bot_mxid}
        headers: dict[str, str] = {"Authorization": f"Bearer {self.az.as_token}"}
        resp_headers = {
            "Content-Security-Policy": (
                "sandbox; default-src 'none'; script-src 'none'; style-src 'none'; object-src 'none';"
            ),
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Methods": "GET, HEAD, OPTIONS",
            "Content-Disposition": "attachment",
        }
        started_writing = False
        try:
            async with self.api.session.get(download_url, params=query_params, headers=headers) as dl_resp:
                resp = web.StreamResponse(status=dl_resp.status, headers=resp_headers)
                if dl_resp.content_length:
                    resp.content_length = dl_resp.content_length
                resp.content_type = dl_resp.content_type
                if "Content-Disposition" in dl_resp.headers:
                    resp.headers["Content-Disposition"] = dl_resp.headers["Content-Disposition"]
                elif resp.status >= 300:
                    del resp.headers["Content-Disposition"]
                started_writing = True
                await resp.prepare(req)
                async for chunk, end_of_chunk in dl_resp.content.iter_chunks():
                    await resp.write(chunk)
                return resp
        except Exception:
            if not started_writing:
                logging.exception("Failed to fetch media")
                return web.Response(status=502, text="Failed to fetch media")

    def mxc_to_url(self, mxc: str, filename=None):
        if not self.media_endpoint:
            return "<media unavailable>"
        try:
            server, media_id = self.api.parse_mxc_uri(mxc)
        except ValueError:
            return "<invalid mxc URI>"

        if filename is None:
            filename = ""
        else:
            filename = "/" + urllib.parse.quote(filename)

        media_path = self.media_path.format(
            server=server, media_id=media_id, checksum=self.mxc_checksum(server, media_id), filename=filename
        )

        return urllib.parse.urljoin(self.media_endpoint, media_path)

    async def reset(self, config_file, homeserver_url):
        with open(config_file) as f:
            registration = yaml.load(f)

        api = HTTPAPI(base_url=homeserver_url, token=registration["as_token"])
        whoami = await api.request(Method.GET, Path.v3.account.whoami)
        self.user_id = whoami["user_id"]
        self.server_name = self.user_id.split(":", 1)[1]
        print("We are " + whoami["user_id"])

        self.az = MauService(
            id=registration["id"],
            domain=self.server_name,
            server=homeserver_url,
            as_token=registration["as_token"],
            hs_token=registration["hs_token"],
            bot_localpart=registration["sender_localpart"],
            state_store=MemoryBridgeStateStore(),
        )

        try:
            await self.az.start(host="127.0.0.1", port=None)
        except Exception:
            logging.exception("Failed to listen.")
            return

        joined_rooms = await self.az.intent.get_joined_rooms()
        print(f"Leaving from {len(joined_rooms)} rooms...")

        for room_id in joined_rooms:
            print(f"Leaving from {room_id}...")
            await self.leave_room(room_id, None)

        print("Resetting configuration...")
        self.config = {}
        await self.save()

        print("All done!")

    def load_reg(self, config_file):
        with open(config_file) as f:
            self.registration = yaml.load(f)

    async def leave_room(self, room_id, members):
        members = members if members else []

        for member in members:
            (name, server) = member.split(":", 1)

            if name.startswith("@" + self.puppet_prefix) and server == self.server_name:
                try:
                    await self.az.intent.user(member).leave_room(room_id)
                except Exception:
                    logging.exception("Removing puppet on leave failed")

        try:
            await self.az.intent.leave_room(room_id)
        except MatrixRequestError:
            pass
        try:
            await self.az.intent.forget_room(room_id)
        except MatrixRequestError:
            pass

    def _keepalive(self):
        async def put_presence():
            try:
                await self.az.intent.set_presence(self.user_id)
            except Exception:
                pass

        asyncio.ensure_future(put_presence())
        asyncio.get_running_loop().call_later(60, self._keepalive)

    async def ensure_hidden_room(self):
        use_hidden_room = self.config.get("use_hidden_room", False)

        if not self.hidden_room and use_hidden_room:
            try:
                resp = await self.az.intent.api.request(Method.GET, Path.v3.capabilities)
                if resp["capabilities"]["m.room_versions"]["available"].get("9", None) == "stable":
                    self.hidden_room = await HiddenRoom.create(self)
                else:
                    m = "No stable version 9 rooms available, hidden room disabled."
                    logging.info(m)
                    raise Exception(m)
            except KeyError:
                m = "Unexpected capabilities response from server."
                logging.debug(m)
                raise Exception(m)
        elif self.hidden_room and not use_hidden_room:
            joined = await self.az.state_store.get_member_profiles(self.hidden_room.id, (Membership.JOIN,))

            self.unregister_room(self.hidden_room.id)
            await self.leave_room(self.hidden_room.id, joined.keys())
            self.hidden_room = None

            for room in self._rooms.values():
                if room.hidden_room_id:
                    # Re-Run post init if room has a hidden room set
                    await room.post_init()

        return use_hidden_room

    async def run(self, listen_address, listen_port, homeserver_url, owner, safe_mode, media_proxy):
        if "sender_localpart" not in self.registration:
            print("Missing sender_localpart from registration file.")
            sys.exit(1)

        if "namespaces" not in self.registration or "users" not in self.registration["namespaces"]:
            print("User namespaces missing from registration file.")
            sys.exit(1)

        # remove self namespace if exists
        ns_users = [
            x
            for x in self.registration["namespaces"]["users"]
            if x["regex"].split(":")[0] != f"@{self.registration['sender_localpart']}"
        ]

        if len(ns_users) != 1:
            print("A single user namespace is required for puppets in the registration file.")
            sys.exit(1)

        if "exclusive" not in ns_users[0] or not ns_users[0]["exclusive"]:
            print("User namespace must be exclusive.")
            sys.exit(1)

        m = re.match(r"^@(.+)([\_/])\.[\*\+]:?", ns_users[0]["regex"])
        if not m:
            print(
                "User namespace regex must be an exact prefix like '@irc_.*' that includes the separator character (_ or /)."
            )
            sys.exit(1)

        self.puppet_separator = m.group(2)
        self.puppet_prefix = m.group(1) + self.puppet_separator

        print(f"Heisenbridge v{__version__}", flush=True)
        if safe_mode:
            print("Safe mode is enabled.", flush=True)
        if media_proxy:
            print("Media proxy only mode.", flush=True)

        url = urllib.parse.urlparse(homeserver_url)
        ws = None
        if url.scheme in ["ws", "wss"]:
            print(
                f"Using websockets to receive transactions. Listening is still enabled on http://{listen_address}:{listen_port}"
            )
            ws = AppserviceWebsocket(homeserver_url, self.registration["as_token"], self._on_mx_event)
            homeserver_url = url._replace(scheme=("https" if url.scheme == "wss" else "http")).geturl()
            print(f"Connecting to HS at {homeserver_url}")

        self.api = HTTPAPI(base_url=homeserver_url, token=self.registration["as_token"])

        # conduit requires that the appservice user is registered before whoami
        wait = 0
        while True:
            try:
                await self.api.request(
                    Method.POST,
                    Path.v3.register,
                    {
                        "type": "m.login.application_service",
                        "username": self.registration["sender_localpart"],
                    },
                )
                logging.debug("Appservice user registration succeeded.")
                break
            except MUserInUse:
                logging.debug("Appservice user is already registered.")
                break
            except MatrixConnectionError as e:
                if wait < 30:
                    wait += 5
                logging.warning(f"Failed to connect to HS: {e}, retrying in {wait} seconds...")
                await asyncio.sleep(wait)
            except Exception:
                logging.exception("Unexpected failure when registering appservice user.")
                sys.exit(1)

        # mautrix migration requires us to call whoami manually at this point
        whoami = await self.api.request(Method.GET, Path.v3.account.whoami)

        logging.info("We are " + whoami["user_id"])

        self.user_id = whoami["user_id"]
        self.server_name = self.user_id.split(":", 1)[1]

        self.az = MauService(
            id=self.registration["id"],
            domain=self.server_name,
            server=homeserver_url,
            as_token=self.registration["as_token"],
            hs_token=self.registration["hs_token"],
            bot_localpart=self.registration["sender_localpart"],
            state_store=MemoryBridgeStateStore(),
        )
        self.az.matrix_event_handler(self._on_mx_event)
        self.az.app.router.add_get("/_heisenbridge/media/{server}/{media_id}/{checksum}/{filename}", self.proxy_media)
        self.az.app.router.add_get("/_heisenbridge/media/{server}/{media_id}/{checksum}", self.proxy_media)

        try:
            await self.az.start(host=listen_address, port=listen_port)
        except Exception:
            logging.exception("Failed to listen.")
            sys.exit(1)

        try:
            await self.az.intent.ensure_registered()
            logging.debug("Appservice user exists at least now.")
        except Exception:
            logging.exception("Unexpected failure when registering appservice user.")
            sys.exit(1)

        if "heisenbridge" in self.registration and "displayname" in self.registration["heisenbridge"]:
            try:
                logging.debug(
                    f"Overriding displayname from registration file to {self.registration['heisenbridge']['displayname']}"
                )
                await self.az.intent.set_displayname(self.registration["heisenbridge"]["displayname"])
            except MatrixRequestError as e:
                logging.warning(f"Failed to set displayname: {str(e)}")

        self._rooms = {}
        self._users = {}
        self.config = {
            "networks": {},
            "owner": None,
            "allow": {},
            "idents": {},
            "member_sync": "half",
            "max_lines": 0,
            "use_pastebin": False,
            "use_reacts": True,
            "media_url": None,
            "media_path": None,
            "media_key": None,
            "namespace": self.puppet_prefix,
        }
        logging.debug(f"Default config: {self.config}")
        self.synapse_admin = False

        try:
            is_admin = await self.api.request(Method.GET, SynapseAdminPath.v1.users[self.user_id].admin)
            self.synapse_admin = is_admin["admin"]
        except MForbidden:
            logging.info(f"We ({self.user_id}) are not a server admin, inviting puppets is required.")
        except Exception:
            logging.info("Seems we are not connected to Synapse, inviting puppets is required.")

        # load config from HS
        await self.load()

        if "heisenbridge" in self.registration and "media_key" in self.registration["heisenbridge"]:
            self.media_key = self.registration["heisenbridge"]["media_key"].encode("utf-8")
        elif self.config["media_key"]:
            self.media_key = self.config["media_key"].encode("utf-8")
        else:
            self.media_key = self.registration["hs_token"].encode("utf-8")

        # use configured media_url for endpoint if we have it
        if "heisenbridge" in self.registration and "media_url" in self.registration["heisenbridge"]:
            logging.debug(
                f"Overriding media URL from registration file to {self.registration['heisenbridge']['media_url']}"
            )
            self.media_endpoint = self.registration["heisenbridge"]["media_url"]
        elif self.config["media_url"]:
            self.media_endpoint = self.config["media_url"]

        # use configured media_path for media_path if we have it
        if "heisenbridge" in self.registration and "media_path" in self.registration["heisenbridge"]:
            logging.debug(
                f"Overriding media path from registration file to {self.registration['heisenbridge']['media_path']}"
            )
            self.media_path = self.registration["heisenbridge"]["media_path"]
        elif self.config["media_path"]:
            self.media_path = self.config["media_path"]
        else:
            self.media_path = self.DEFAULT_MEDIA_PATH

        if media_proxy:
            logging.info("Media proxy mode startup complete")
            await asyncio.Event().wait()
            return

        logging.info("Starting presence loop")
        self._keepalive()

        # do a little migration for servers, remove this later
        for network in self.config["networks"].values():
            new_servers = []

            for server in network["servers"]:
                if isinstance(server, str):
                    new_servers.append({"address": server, "port": 6667, "tls": False})

            if len(new_servers) > 0:
                logging.debug("Migrating servers from old to new config format")
                network["servers"] = new_servers

        logging.debug(f"Merged configuration from HS: {self.config}")

        # prevent starting bridge with changed namespace
        if self.config["namespace"] != self.puppet_prefix:
            logging.error(
                f"Previously used namespace '{self.config['namespace']}' does not match current '{self.puppet_prefix}'."
            )
            sys.exit(1)

        # honor command line owner
        if owner is not None and self.config["owner"] != owner:
            logging.info(f"Overriding loaded owner with '{owner}'")
            self.config["owner"] = owner

        # always ensure our merged and migrated configuration is up-to-date
        await self.save()

        print("Fetching joined rooms...", flush=True)

        joined_rooms = await self.az.intent.get_joined_rooms()
        logging.debug(f"Appservice rooms: {joined_rooms}")

        print(f"Bridge is in {len(joined_rooms)} rooms, initializing them...", flush=True)

        Room.init_class(self.az)
        self.hidden_room = None

        # room types and their init order, network must be before chat and group
        room_types = [HiddenRoom, ControlRoom, NetworkRoom, PrivateRoom, ChannelRoom, PlumbedRoom, SpaceRoom]

        room_type_map = {}
        for room_type in room_types:
            room_type.init_class(self.az)
            room_type_map[room_type.__name__] = room_type

        # we always auto-open control room for owner
        owner_control_open = False

        # import all rooms
        for room_id in joined_rooms:
            joined = {}

            try:
                config = await self.az.intent.get_account_data("irc", room_id)

                if "type" not in config or "user_id" not in config:
                    raise Exception("Invalid config")

                cls = room_type_map.get(config["type"])
                if not cls:
                    raise Exception("Unknown room type")

                # refresh room members state
                await self.az.intent.get_room_members(room_id)

                joined = await self.az.state_store.get_member_profiles(room_id, (Membership.JOIN,))
                banned = await self.az.state_store.get_members(room_id, (Membership.BAN,))

                room = cls(id=room_id, user_id=config["user_id"], serv=self, members=joined.keys(), bans=banned)
                room.from_config(config)

                join_rules = await self.az.intent.get_state_event(room_id, EventType.ROOM_JOIN_RULES)
                if join_rules.join_rule == JoinRule.RESTRICTED and join_rules.allow:
                    room.hidden_room_id = join_rules.allow[0].room_id

                # add to room displayname
                for user_id, member in joined.items():
                    if member.displayname is not None:
                        room.displaynames[user_id] = member.displayname
                    # add to global puppet cache if it's a puppet
                    if user_id.startswith("@" + self.puppet_prefix) and self.is_local(user_id):
                        self._users[user_id] = member.displayname

                # only add valid rooms to event handler
                if room.is_valid():
                    self._rooms[room_id] = room
                else:
                    room.cleanup()
                    raise Exception("Room validation failed after init")

                if cls is HiddenRoom:
                    self.hidden_room = room

                if cls == ControlRoom and room.user_id == self.config["owner"]:
                    owner_control_open = True
            except Exception:
                logging.exception(f"Failed to reconfigure room {room_id} during init, leaving.")

                # regardless of same mode, we ignore this room
                self.unregister_room(room_id)

                if safe_mode:
                    print("Safe mode enabled, not leaving room.", flush=True)
                else:
                    await self.leave_room(room_id, joined.keys())

        try:
            await self.ensure_hidden_room()
        except Exception as e:
            logging.debug(f"Failed setting up hidden room: {e}")

        print("All valid rooms initialized, connecting network rooms...", flush=True)

        wait = 1
        for room in list(self._rooms.values()):
            await room.post_init()

            # check again if we're still valid
            if not room.is_valid():
                logging.debug(f"Room {room.id} failed validation after post init, leaving.")

                self.unregister_room(room.id)

                if not safe_mode:
                    await self.leave_room(room.id, room.members)

                continue

            # connect network rooms one by one, this may take a while
            if type(room) == NetworkRoom and room.connected:

                def sync_connect(room):
                    asyncio.ensure_future(room.connect())

                asyncio.get_running_loop().call_later(wait, sync_connect, room)
                wait += 1

        print(f"Init done with {wait-1} networks connecting, bridge is now running!", flush=True)

        await self.push_bridge_state(BridgeStateEvent.UNCONFIGURED)

        # late start WS to avoid getting transactions too early
        if ws:
            await ws.start()

        if self.config["owner"] and not owner_control_open:
            print(f"Opening control room for owner {self.config['owner']}")
            try:
                room_id = await self.az.intent.create_room(
                    invitees=[self.config["owner"]], custom_request_fields={"com.beeper.auto_join_invites": True}
                )

                room = ControlRoom(
                    id=room_id, user_id=self.config["owner"], serv=self, members=[self.config["owner"]], bans=[]
                )
                await room.save()
                self.register_room(room)

                await self.az.intent.join_room(room.id)

                # show help on open
                await room.show_help()
            except Exception:
                print("Failed to create control room, huh")

        await asyncio.Event().wait()


async def async_main():
    parser = argparse.ArgumentParser(
        prog=os.path.basename(sys.executable) + " -m " + __package__,
        description=f"a bouncer-style Matrix IRC bridge (v{__version__})",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-v", "--verbose", help="logging verbosity level: once is info, twice is debug", action="count", default=0
    )
    req = parser.add_mutually_exclusive_group(required=True)
    req.add_argument(
        "-c",
        "--config",
        help="registration YAML file path, must be writable if generating",
    )
    req.add_argument(
        "--version",
        action="store_true",
        help="show bridge version",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-l",
        "--listen-address",
        help="bridge listen address (default: as specified in url in config, 127.0.0.1 otherwise)",
    )
    parser.add_argument(
        "-p",
        "--listen-port",
        help="bridge listen port (default: as specified in url in config, 9898 otherwise)",
        type=int,
    )
    parser.add_argument("-u", "--uid", help="user id to run as", default=None)
    parser.add_argument("-g", "--gid", help="group id to run as", default=None)
    parser.add_argument("-i", "--identd", action="store_true", help="enable identd service")
    parser.add_argument("--identd-port", type=int, default="113", help="identd listen port")
    parser.add_argument(
        "--generate",
        action="store_true",
        help="generate registration YAML for Matrix homeserver (Synapse)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--generate-compat",
        action="store_true",
        help="generate registration YAML for Matrix homeserver (Dendrite and Conduit)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--reset",
        action="store_true",
        help="reset ALL bridge configuration from homeserver and exit",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--media-proxy",
        action="store_true",
        help="run in media proxy mode",
        default=False,
    )
    parser.add_argument(
        "--safe-mode",
        action="store_true",
        help="prevent appservice from leaving invalid rooms on startup (for debugging)",
    )
    parser.add_argument(
        "-o",
        "--owner",
        help="set owner MXID (eg: @user:homeserver) or first talking local user will claim the bridge",
        default=None,
    )
    parser.add_argument(
        "homeserver",
        nargs="?",
        help="URL of Matrix homeserver",
        default="http://localhost:8008",
    )

    args = parser.parse_args()

    logging_level = logging.WARNING
    if args.verbose > 0:
        logging_level = logging.INFO
        if args.verbose > 1:
            logging_level = logging.DEBUG

    logging.basicConfig(stream=sys.stdout, level=logging_level)

    if "generate" in args or "generate_compat" in args:
        letters = string.ascii_letters + string.digits

        registration = {
            "id": "heisenbridge",
            "url": "http://{}:{}".format(args.listen_address or "127.0.0.1", args.listen_port or 9898),
            "as_token": "".join(random.choice(letters) for i in range(64)),
            "hs_token": "".join(random.choice(letters) for i in range(64)),
            "rate_limited": False,
            "sender_localpart": "heisenbridge",
            "namespaces": {
                "users": [{"regex": "@irc_.*", "exclusive": True}],
                "aliases": [],
                "rooms": [],
            },
        }

        if "generate_compat" in args:
            registration["namespaces"]["users"].append({"regex": "@heisenbridge:.*", "exclusive": True})

        if os.path.isfile(args.config):
            print("Registration file already exists, not overwriting.")
            sys.exit(1)

        if args.config == "-":
            yaml.dump(registration, sys.stdout)
        else:
            with open(args.config, "w") as f:
                yaml.dump(registration, f)

            print(f"Registration file generated and saved to {args.config}")
    elif "reset" in args:
        service = BridgeAppService()
        await service.reset(args.config, args.homeserver)
    elif "version" in args:
        print(__version__)
    else:
        service = BridgeAppService()
        identd = None

        service.load_reg(args.config)

        if args.identd and not args.media_proxy:
            identd = Identd()
            await identd.start_listening(service, args.identd_port)

        if os.getuid() == 0:
            if args.gid:
                gid = grp.getgrnam(args.gid).gr_gid
                os.setgid(gid)
                os.setgroups([])

            if args.uid:
                uid = pwd.getpwnam(args.uid).pw_uid
                os.setuid(uid)

        os.umask(0o077)

        listen_address = args.listen_address
        listen_port = args.listen_port

        if not listen_address:
            listen_address = "127.0.0.1"

            try:
                url = urllib.parse.urlparse(service.registration["url"])
                if url.hostname:
                    listen_address = url.hostname
            except Exception:
                pass

        if not listen_port:
            listen_port = 9898

            try:
                url = urllib.parse.urlparse(service.registration["url"])
                if url.port:
                    listen_port = url.port
            except Exception:
                pass

        await service.run(listen_address, listen_port, args.homeserver, args.owner, args.safe_mode, args.media_proxy)


def main():
    asyncio.run(async_main())


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import logging
from abc import ABC
from abc import abstractmethod
from typing import List

from mautrix.api import Method
from mautrix.api import Path
from mautrix.errors import MNotFound


class Room:
    pass


class AppService(ABC):
    user_id: str
    server_name: str
    config: dict
    hidden_room: Room

    async def load(self):
        try:
            self.config.update(await self.az.intent.get_account_data("irc"))
        except MNotFound:
            await self.save()

    async def save(self):
        await self.az.intent.set_account_data("irc", self.config)

    async def create_room(self, name: str, topic: str, invite: List[str], restricted: str = None) -> str:
        req = {
            "visibility": "private",
            "name": name,
            "topic": topic,
            "invite": invite,
            "is_direct": False,
            "power_level_content_override": {
                "users_default": 0,
                "invite": 100,
                "kick": 100,
                "redact": 100,
                "ban": 100,
                "events": {
                    "m.room.name": 0,
                    "m.room.avatar": 0,  # these work as long as rooms are private
                },
            },
            "com.beeper.auto_join_invites": True,
        }

        if restricted is not None:
            resp = await self.az.intent.api.request(Method.GET, Path.v3.capabilities)
            try:
                def_ver = resp["capabilities"]["m.room_versions"]["default"]
            except KeyError:
                logging.debug("Unexpected capabilities reply")
                def_ver = None

            # If room version is in range of 1..8, request v9
            if def_ver in [str(v) for v in range(1, 9)]:
                req["room_version"] = "9"

            req["initial_state"] = [
                {
                    "type": "m.room.join_rules",
                    "state_key": "",
                    "content": {
                        "join_rule": "restricted",
                        "allow": [{"type": "m.room_membership", "room_id": restricted}],
                    },
                }
            ]

        resp = await self.az.intent.api.request(Method.POST, Path.v3.createRoom, req)

        return resp["room_id"]

    @abstractmethod
    def register_room(self, room: Room):
        pass

    @abstractmethod
    def find_rooms(self, type=None, user_id: str = None) -> List[Room]:
        pass



# Included to allow for editable installs
import importlib.util

from setuptools import Command
from setuptools import setup
from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist

# pull git or local version
spec = importlib.util.spec_from_file_location("version", "heisenbridge/version.py")
version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version)


class GenerateVersionCommand(Command):
    description = "Generate version.txt"
    user_options = []

    def run(self):
        with open("heisenbridge/version.txt", "w") as version_file:
            version_file.write(version.__version__)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class BuildPyCommand(build_py):
    def run(self):
        self.run_command("gen_version")
        build_py.run(self)


class SDistCommand(sdist):
    def run(self):
        self.run_command("gen_version")
        sdist.run(self)


setup(
    version=version.__version__,
    cmdclass={"gen_version": GenerateVersionCommand, "build_py": BuildPyCommand, "sdist": SDistCommand},
    packages=["heisenbridge"],
    package_data={"heisenbridge": ["version.txt"]},
)



import asyncio
import html
import logging
from typing import Dict
from typing import List
from typing import Optional

from irc.modes import parse_channel_modes

from heisenbridge.command_parse import CommandParser
from heisenbridge.private_room import parse_irc_formatting
from heisenbridge.private_room import PrivateRoom
from heisenbridge.private_room import unix_to_local


class NetworkRoom:
    pass


class ChannelRoom(PrivateRoom):
    key: Optional[str]
    member_sync: str
    autocmd: str
    names_buffer: List[str]
    bans_buffer: List[str]
    on_channel: List[str]

    def init(self) -> None:
        super().init()

        self.key = None
        self.autocmd = None

        # for migration the class default is full
        self.member_sync = "full"

        cmd = CommandParser(
            prog="AUTOCMD",
            description="run commands on join",
            epilog=(
                "Works _exactly_ like network AUTOCMD and runs in the network context."
                " You can use this to login to bots or other services after joining a channel."
            ),
        )
        cmd.add_argument("command", nargs="*", help="commands separated with ';'")
        cmd.add_argument("--remove", action="store_true", help="remove stored command")
        self.commands.register(cmd, self.cmd_autocmd)

        cmd = CommandParser(
            prog="SYNC",
            description="override IRC member sync type for this room",
            epilog="Note: To force full sync after setting to full, use the NAMES command",
        )
        group = cmd.add_mutually_exclusive_group()
        group.add_argument("--lazy", help="set lazy sync, members are added when they talk", action="store_true")
        group.add_argument(
            "--half", help="set half sync, members are added when they join or talk", action="store_true"
        )
        group.add_argument("--full", help="set full sync, members are fully synchronized", action="store_true")
        group.add_argument(
            "--off",
            help="disable member sync completely, the bridge will relay all messages, may be useful during spam attacks",
            action="store_true",
        )
        self.commands.register(cmd, self.cmd_sync)

        cmd = CommandParser(
            prog="MODE",
            description="send MODE command",
            epilog=(
                "Can be used to change channel modes, ban lists or invoke/manage custom lists.\n"
                "It is very network specific what modes or lists are supported, please see their documentation"
                " for comprehensive help.\n"
                "\n"
                "Note: Some common modes and lists may have a command, see HELP.\n"
            ),
        )
        cmd.add_argument("args", nargs="*", help="MODE command arguments")
        self.commands.register(cmd, self.cmd_mode)

        cmd = CommandParser(
            prog="NAMES",
            description="list channel members",
            epilog=(
                "Sends a NAMES command to server.\n"
                "\n"
                "This can be used to see what IRC permissions users currently have on this channel.\n"
                "\n"
                "Note: In addition this will resynchronize the Matrix room members list and may cause joins/leaves"
                " if it has fallen out of sync.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_names)

        # plumbs have a slightly adjusted version
        if type(self) == ChannelRoom:
            cmd = CommandParser(prog="TOPIC", description="show or set channel topic")
            cmd.add_argument("text", nargs="*", help="topic text if setting")
            self.commands.register(cmd, self.cmd_topic)

        cmd = CommandParser(prog="BANS", description="show channel ban list")
        self.commands.register(cmd, self.cmd_bans)

        cmd = CommandParser(prog="OP", description="op someone")
        cmd.add_argument("nick", help="nick to target")
        self.commands.register(cmd, self.cmd_op)

        cmd = CommandParser(prog="DEOP", description="deop someone")
        cmd.add_argument("nick", help="nick to target")
        self.commands.register(cmd, self.cmd_deop)

        cmd = CommandParser(prog="VOICE", description="voice someone")
        cmd.add_argument("nick", help="nick to target")
        self.commands.register(cmd, self.cmd_voice)

        cmd = CommandParser(prog="DEVOICE", description="devoice someone")
        cmd.add_argument("nick", help="nick to target")
        self.commands.register(cmd, self.cmd_devoice)

        cmd = CommandParser(prog="KICK", description="kick someone")
        cmd.add_argument("nick", help="nick to target")
        cmd.add_argument("reason", nargs="*", help="reason")
        self.commands.register(cmd, self.cmd_kick)

        cmd = CommandParser(prog="KB", description="kick and ban someone")
        cmd.add_argument("nick", help="nick to target")
        cmd.add_argument("reason", nargs="*", help="reason")
        self.commands.register(cmd, self.cmd_kb)

        cmd = CommandParser(prog="JOIN", description="join this channel if not on it")
        self.commands.register(cmd, self.cmd_join)

        cmd = CommandParser(prog="PART", description="leave this channel temporarily")
        self.commands.register(cmd, self.cmd_part)

        cmd = CommandParser(
            prog="STOP",
            description="immediately clear all queued IRC events like long messages",
            epilog="Use this to stop accidental long pastes, also known as STAHP!",
        )
        self.commands.register(cmd, self.cmd_stop, ["STOP!", "STAHP", "STAHP!"])

        cmd = CommandParser(
            prog="UPGRADE",
            description="Perform any potential bridge-side upgrades of the room",
        )
        cmd.add_argument("--undo", action="store_true", help="undo previously performed upgrade")
        self.commands.register(cmd, self.cmd_upgrade)

        self.names_buffer = []
        self.bans_buffer = []
        self.on_channel = []

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "key" in config:
            self.key = config["key"]

        if "member_sync" in config:
            self.member_sync = config["member_sync"]

        if "autocmd" in config:
            self.autocmd = config["autocmd"]

        # initialize lazy members dict if sync is not off
        if self.member_sync != "off":
            if self.lazy_members is None:
                self.lazy_members = {}
        else:
            self.lazy_members = None

    def to_config(self) -> dict:
        return {**(super().to_config()), "key": self.key, "member_sync": self.member_sync, "autocmd": self.autocmd}

    @staticmethod
    def create(network: NetworkRoom, name: str) -> "ChannelRoom":
        logging.debug(f"ChannelRoom.create(network='{network.name}', name='{name}'")

        room = ChannelRoom(None, network.user_id, network.serv, [network.user_id, network.serv.user_id], [])
        room.name = name.lower()
        room.network = network
        room.network_id = network.id
        room.network_name = network.name

        # fetch stored channel key if used for join command
        if room.name in network.keys:
            room.key = network.keys[room.name]
            del network.keys[room.name]

        # stamp global member sync setting at room creation time
        room.member_sync = network.serv.config["member_sync"]

        room.max_lines = network.serv.config["max_lines"]
        room.use_pastebin = network.serv.config["use_pastebin"]
        room.use_reacts = network.serv.config["use_reacts"]

        asyncio.ensure_future(room._create_mx(name))
        return room

    async def _create_mx(self, name):
        # handle !room names properly
        visible_name = name
        if visible_name.startswith("!"):
            visible_name = "!" + visible_name[6:]

        if self.serv.hidden_room:
            self.hidden_room_id = self.serv.hidden_room.id

        self.id = await self.network.serv.create_room(
            f"{visible_name} ({self.network.name})",
            "",
            [self.network.user_id],
            self.hidden_room_id,
        )
        self.serv.register_room(self)
        await self.save()
        # start event queue now that we have an id
        self._queue.start()

        # attach to network space
        if self.network.space:
            await self.network.space.attach(self.id)

    def is_valid(self) -> bool:
        if not self.in_room(self.user_id):
            return False

        return super().is_valid()

    def cleanup(self) -> None:
        if self.network:
            if self.network.conn and self.network.conn.connected:
                self.network.conn.part(self.name)

        super().cleanup()

    async def cmd_autocmd(self, args) -> None:
        autocmd = " ".join(args.command)

        if args.remove:
            self.autocmd = None
            await self.save()
            self.send_notice("Autocmd removed.", forward=args._forward)
            return

        if autocmd == "":
            self.send_notice(f"Configured autocmd: {self.autocmd if self.autocmd else ''}", forward=args._forward)
            return

        self.autocmd = autocmd
        await self.save()
        self.send_notice(f"Autocmd set to {self.autocmd}", forward=args._forward)

    async def cmd_sync(self, args):
        if args.lazy:
            self.member_sync = "lazy"
            await self.save()
        elif args.half:
            self.member_sync = "half"
            await self.save()
        elif args.full:
            self.member_sync = "full"
            await self.save()
        elif args.off:
            self.member_sync = "off"
            # prevent anyone already in lazy list to be invited
            self.lazy_members = None
            await self.save()

        self.send_notice(f"Member sync is set to {self.member_sync}", forward=args._forward)

    async def cmd_mode(self, args) -> None:
        self.network.conn.mode(self.name, " ".join(args.args))

    async def cmd_modes(self, args) -> None:
        self.network.conn.mode(self.name, "")

    async def cmd_names(self, args) -> None:
        self.network.conn.names(self.name)

    async def cmd_bans(self, args) -> None:
        self.network.conn.mode(self.name, "+b")

    async def cmd_op(self, args) -> None:
        self.network.conn.mode(self.name, f"+o {args.nick}")

    async def cmd_deop(self, args) -> None:
        self.network.conn.mode(self.name, f"-o {args.nick}")

    async def cmd_voice(self, args) -> None:
        self.network.conn.mode(self.name, f"+v {args.nick}")

    async def cmd_devoice(self, args) -> None:
        self.network.conn.mode(self.name, f"-v {args.nick}")

    async def cmd_topic(self, args) -> None:
        self.network.conn.topic(self.name, " ".join(args.text))

    async def cmd_kick(self, args) -> None:
        self.network.conn.kick(self.name, args.nick, " ".join(args.reason))

    async def cmd_kb(self, args) -> None:
        self.network.kickban(self.name, args.nick, " ".join(args.reason))

    async def cmd_join(self, args) -> None:
        self.network.conn.join(self.name, self.key)

    async def cmd_part(self, args) -> None:
        self.network.conn.part(self.name)

    async def cmd_stop(self, args) -> None:
        filtered = self.network.conn.remove_tag(self.name)
        self.send_notice(f"{filtered} messages removed from queue.")

    def on_pubmsg(self, conn, event):
        self.on_privmsg(conn, event)

    def on_pubnotice(self, conn, event):
        self.on_privnotice(conn, event)

    def on_namreply(self, conn, event) -> None:
        self.names_buffer.extend(event.arguments[2].split())

    def _add_puppet(self, nick):
        irc_user_id = self.serv.irc_user_id(self.network.name, nick)

        self.ensure_irc_user_id(self.network.name, nick)
        self.join(irc_user_id, nick)

    def _remove_puppet(self, user_id, reason=None):
        if user_id == self.serv.user_id or user_id == self.user_id:
            return

        self.leave(user_id, reason)

    def on_endofnames(self, conn, event) -> None:
        to_remove = []
        to_add = []
        names = list(self.names_buffer)
        self.names_buffer = []
        modes: Dict[str, List[str]] = {}
        others = []
        on_channel = []

        # always reset lazy list because it can be toggled on-the-fly
        self.lazy_members = {} if self.member_sync != "off" else None

        # build to_remove list from our own puppets
        for member in self.members:
            (name, server) = member.split(":", 1)

            if name.startswith("@" + self.serv.puppet_prefix) and server == self.serv.server_name:
                to_remove.append(member)

        for nick in names:
            nick, mode = self.serv.strip_nick(nick)
            on_channel.append(nick.lower())

            if mode:
                if mode not in modes:
                    modes[mode] = []

                if nick == conn.real_nickname:
                    modes[mode].append(nick + " (you)")
                else:
                    modes[mode].append(nick)
            else:
                if nick == conn.real_nickname:
                    others.append(nick + " (you)")
                else:
                    others.append(nick)

            # convert to mx id, check if we already have them
            irc_user_id = self.serv.irc_user_id(self.network.name, nick)

            # make sure this user is not removed from room
            if irc_user_id in to_remove:
                to_remove.remove(irc_user_id)
                continue

            # ignore adding us here, only lazy join on echo allowed
            if nick == conn.real_nickname:
                continue

            # if this user is not in room, add to invite list
            if not self.in_room(irc_user_id):
                to_add.append((irc_user_id, nick))

            # always put everyone in the room to lazy list if we have any member sync
            if self.lazy_members is not None:
                self.lazy_members[irc_user_id] = nick

        # never remove us or appservice
        if self.serv.user_id in to_remove:
            to_remove.remove(self.serv.user_id)
        if self.user_id in to_remove:
            to_remove.remove(self.user_id)

        self.send_notice(
            "Synchronizing members:"
            + f" got {len(names)} from server,"
            + f" {len(self.members)} in room,"
            + f" {len(to_add)} will be invited and {len(to_remove)} removed."
        )

        # known common mode names
        modenames = {
            "~": "owner",
            "&": "admin",
            "@": "op",
            "%": "half-op",
            "+": "voice",
        }

        # show modes from top to bottom
        for mode, name in modenames.items():
            if mode in modes:
                nicks = sorted(modes[mode], key=str.casefold)
                self.send_notice(f"Users with {name} ({mode}): {', '.join(nicks)}")
                del modes[mode]

        # show unknown modes
        for mode, nicks in modes.items():
            nicks = sorted(nicks, key=str.casefold)
            self.send_notice(f"Users with '{mode}': {', '.join(nicks)}")

        # show everyone else
        if len(others) > 0:
            others = sorted(others, key=str.casefold)
            self.send_notice(f"Users: {', '.join(others)}")

        if self.member_sync == "full":
            for irc_user_id, nick in to_add:
                self._add_puppet(nick)
        else:
            self.send_notice(f"Member sync is set to {self.member_sync}, skipping invites.")

        for irc_user_id in to_remove:
            self._remove_puppet(irc_user_id)

        # trust the names reply is always up-to-date
        self.on_channel = on_channel

    def is_on_channel(self, nick):
        return nick.lower() in self.on_channel

    def channel_join(self, nick):
        nick = nick.lower()
        if nick not in self.on_channel:
            self.on_channel.append(nick)

    def channel_leave(self, nick):
        nick = nick.lower()
        if nick in self.on_channel:
            self.on_channel.remove(nick)

    def on_join(self, conn, event) -> None:
        self.channel_join(event.source.nick)

        # we don't need to sync ourself
        if conn.real_nickname == event.source.nick:
            self.send_notice(f"Joined {event.target} as {event.source.nick} ({event.source.userhost})")

            # sync channel modes/key on join
            self.network.conn.mode(self.name, "")

            # send autocmd if we have one
            if self.autocmd:

                async def autocmd(self):
                    self.send_notice("Executing channel autocmd.")
                    try:
                        await self.network.commands.trigger(
                            self.autocmd, allowed=["RAW", "MSG", "NICKSERV", "NS", "CHANSERV", "CS", "WAIT"]
                        )
                    except Exception as e:
                        self.send_notice(f"Channel autocmd failed: {str(e)}")

                asyncio.ensure_future(autocmd(self))

            return

        # ensure, append, invite and join
        if self.member_sync == "full" or self.member_sync == "half":
            self._add_puppet(event.source.nick)
        elif self.member_sync != "off":
            irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)
            self.join(irc_user_id, event.source.nick, lazy=True)

    def on_part(self, conn, event) -> None:
        self.channel_leave(event.source.nick)

        # we don't need to sync ourself
        if conn.real_nickname == event.source.nick:
            # immediately dequeue all future events
            conn.remove_tag(event.target.lower())

            self.send_notice_html(
                f"You left the channel. To rejoin, type <b>JOIN {event.target}</b> in the <b>{self.network.name}</b> network room."
            )
            self.send_notice_html("If you want to permanently leave you need to leave this room.")
            return

        irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)
        self._remove_puppet(irc_user_id, event.arguments[0] if len(event.arguments) else None)

    def on_quit(self, conn, event) -> None:
        self.channel_leave(event.source.nick)
        irc_user_id = self.serv.irc_user_id(self.network.name, event.source.nick)
        self._remove_puppet(irc_user_id, f"Quit: {event.arguments[0]}")

    def update_key(self, modes):
        for sign, key, value in parse_channel_modes(" ".join(modes)):
            # update channel key
            if key == "k":
                value = None if sign == "-" else value
                if value != self.key:
                    self.key = value
                    if self.id is not None:
                        asyncio.ensure_future(self.save())

    def on_badchannelkey(self, conn, event) -> None:
        self.send_notice(event.arguments[1] if len(event.arguments) > 1 else "Incorrect channel key, join failed.")
        self.send_notice_html(
            f"Use <b>JOIN {html.escape(event.arguments[0])} &lt;key&gt;</b> in the network room to rejoin this channel."
        )

    def on_chanoprivsneeded(self, conn, event) -> None:
        self.send_notice(event.arguments[1] if len(event.arguments) > 1 else "You're not operator.")

    def on_cannotsendtochan(self, conn, event) -> None:
        self.send_notice(event.arguments[1] if len(event.arguments) > 1 else "Cannot send to channel.")

    def on_mode(self, conn, event) -> None:
        modes = list(event.arguments)

        self.send_notice("{} set modes {}".format(event.source.nick, " ".join(modes)))
        self.update_key(modes)

    def on_notopic(self, conn, event) -> None:
        self.send_notice(event.arguments[1] if len(event.arguments) > 1 else "No topic is set.")
        self.set_topic("")

    def on_currenttopic(self, conn, event) -> None:
        (plain, formatted) = parse_irc_formatting(event.arguments[1])
        self.send_notice(f"Topic is '{plain}'")
        self.set_topic(plain)

    def on_topicinfo(self, conn, event) -> None:
        settime = unix_to_local(event.arguments[2]) if len(event.arguments) > 2 else "?"
        (plain, formatted) = parse_irc_formatting(event.arguments[1])
        self.send_notice(f"Topic set by {plain} at {settime}")

    def on_topic(self, conn, event) -> None:
        self.send_notice("{} changed the topic".format(event.source.nick))
        (plain, formatted) = parse_irc_formatting(event.arguments[0])
        self.set_topic(plain)

    def on_kick(self, conn, event) -> None:
        self.channel_leave(event.arguments[0])

        reason = (": " + event.arguments[1]) if len(event.arguments) > 1 and len(event.arguments[1]) > 0 else ""

        if event.arguments[0] == conn.real_nickname:
            # immediately dequeue all future events
            conn.remove_tag(event.target.lower())

            self.send_notice_html(f"You were kicked from the channel by <b>{event.source.nick}</b>{reason}")
            if self.network.rejoin_kick:
                self.send_notice("Rejoin on kick is enabled, trying to join back immediately...")
                conn.join(event.target)
            else:
                self.send_notice_html(
                    f"To rejoin the channel, type <b>JOIN {event.target}</b> in the <b>{self.network.name}</b> network room."
                )
        else:
            target_user_id = self.serv.irc_user_id(self.network.name, event.arguments[0])
            self.kick(target_user_id, f"Kicked by {event.source.nick}{reason}")

    def on_banlist(self, conn, event) -> None:
        parts = list(event.arguments)
        parts.pop(0)
        self.bans_buffer.append(parts)

    def on_endofbanlist(self, conn, event) -> None:
        bans = self.bans_buffer
        self.bans_buffer = []

        self.send_notice("Current channel bans:")
        for ban in bans:
            strban = f"\t{ban[0]}"

            # all other argumenta are optional
            if len(ban) > 1:
                strban += f" set by {ban[1]}"
            if len(ban) > 2:
                strban += f" at {unix_to_local(ban[2])}"

            self.send_notice(strban)

    def on_channelmodeis(self, conn, event) -> None:
        modes = list(event.arguments)
        modes.pop(0)

        self.send_notice(f"Current channel modes: {' '.join(modes)}")

    def on_channelcreate(self, conn, event) -> None:
        created = unix_to_local(event.arguments[1])
        self.send_notice(f"Channel was created at {created}")

    def on_328(self, conn, event) -> None:
        (plain, formatted) = parse_irc_formatting(event.arguments[1])
        self.send_notice(f"URL for {event.arguments[0]}: {plain}")



import os
import shutil
import subprocess

module_dir = os.path.dirname(__file__)
root_dir = module_dir + "/../"

__version__ = "0.0.0"
__git_version__ = None

if os.path.exists(module_dir + "/version.txt"):
    __version__ = open(module_dir + "/version.txt").read().strip()

if os.path.exists(root_dir + ".git") and shutil.which("git"):
    try:
        git_env = {
            "PATH": os.environ["PATH"],
            "HOME": os.environ["HOME"],
            "LANG": "C",
            "LC_ALL": "C",
        }
        git_bits = (
            subprocess.check_output(["git", "describe", "--tags"], stderr=subprocess.DEVNULL, cwd=root_dir, env=git_env)
            .strip()
            .decode("ascii")
            .split("-")
        )

        __git_version__ = git_bits[0][1:]

        if len(git_bits) > 1:
            __git_version__ += f".dev{git_bits[1]}"

        if len(git_bits) > 2:
            __git_version__ += f"+{git_bits[2]}"

        # always override version with git version if we have a valid version number
        __version__ = __git_version__
    except (subprocess.SubprocessError, OSError):
        pass
